# This open source software code was developed in part or in whole in
# the Human Brain Project, funded from the European Union’s Horizon
# 2020 Framework Programme for Research and Innovation under Specific
# Grant Agreements No. 720270 and No. 785907 (Human Brain Project SGA1
# and SGA2).
#
# etc...
#

import numexpr
import numpy as np

from snudda.utils.snudda_path import snudda_parse_path


# Use np.random.choice for selecting


class NeuronMorphology:
    """ Neuron morphology object. Also see NeuronPrototype class which handles multiple morphology variations. """

    # axonStumpIDFlag should be True if running Network_simulate.py
    # it should be False if we are running Neurodamus simulation.

    def __init__(self,
                 name=None,
                 position=np.zeros((1, 3)),
                 rotation=None,  # np.eye(3),
                 swc_filename=None,
                 param_filename=None,
                 param_data=None,  # TODO: Can we remove this?
                 mech_filename=None,
                 modulation=None,
                 neuron_path=None,
                 morphology_id=None,
                 parameter_id=None,
                 modulation_id=None,
                 parameter_key=None,
                 morphology_key=None,
                 modulation_key=None,
                 verbose=False,
                 load_morphology=True,
                 hoc=None,
                 colour=None,
                 use_cache=True,
                 pickle_version=-1,
                 logfile=None,
                 virtual_neuron=False,
                 axon_stump_id_flag=False,
                 snudda_data=None):

        self.cache_version = 0.99
        self.snudda_data = snudda_data

        self.position = np.array(position)

        if rotation is not None:
            self.rotation = np.array(rotation)
        else:
            self.rotation = None

        self.soma = np.array([])
        self.axon = np.array([])
        self.dend = np.array([])  # 0,1,2: x,y,z  3: radie, 4: dist to soma (all in meters)

        self.axon_density_type = None
        self.dend_density = None
        self.axon_density = None
        self.axon_density_bounds_xyz = None

        self.voxel_size = 5e6
        self.density_bin_size = 10e-6

        self.load_morphology = load_morphology

        # This tells if axon is indexed fully or if only as a stump
        # this affects sectionID for both axon and dendrites
        self.axon_stump_id_flag = axon_stump_id_flag

        # Meta data
        self.name = name
        self.neuron_path = neuron_path
        self.swc_filename = swc_filename
        self.param_filename = param_filename
        self.param_data = param_data
        self.mech_filename = mech_filename
        self.modulation = modulation

        self.morphology_id = morphology_id
        self.parameter_id = parameter_id
        self.modulation_id = modulation_id

        self.parameter_key = parameter_key
        self.morphology_key = morphology_key
        self.modulation_key = modulation_key

        self.verbose = verbose
        self.use_cache = use_cache
        self.pickle_version = pickle_version
        self.log_file = logfile
        self.virtual_neuron = virtual_neuron

        self.rotated_flag = False

        if swc_filename:
            self.cache_filename = swc_filename.replace('.swc', '-cache.pickle')
            assert (self.cache_filename != swc_filename), f"Cached filename: {self.cache_filename} != {swc_filename}"
        else:
            self.cache_filename = None

        # This is used for Neurodamus, which instantiates through hoc files
        if hoc is None:
            hoc = ""

        self.hoc = hoc

        # This is useful when determining connectivity, to exclude pairs
        self.max_axon_radius = 0
        self.max_dend_radius = 0

        # Telling how the different points link together into lines
        self.axon_links = np.array((2, 0))  # These should never be changed after CLONE
        self.dend_links = np.array((2, 0))

        self.dend_sec_id = np.array((1,))
        self.dend_sec_x = np.array((2, 0))

        # Lookup for section length for each section, based on sec_id
        self.sec_id_to_len = None

        # Lookup for index into self.dend_links based on sec_id
        self.sec_id_links = dict()
        self.sec_id_links_x = dict()

        if colour is None:
            self.colour = np.random.random((3,))
        else:
            self.colour = colour

        if load_morphology and self.swc_filename:
            # This loads, rotates and places neuron
            self.load_neuron_morphology()

    ############################################################################

    def load_neuron_morphology(self):

        """ Loads neuron morphology from SWC file or from cache file. """

        if self.use_cache:
            if self.cache_exist():
                # Load existing cache
                try:
                    self.load_cache()
                except Exception as e:

                    # import traceback
                    # tstr = traceback.format_exc()
                    # print(tstr)

                    self.write_log(f"Failed to read cache file, loading: {self.swc_filename}")
                    self.load_swc()
                    self.save_cache()

            else:
                self.write_log("No cache found, create it.")
                # Load SWC and save cache file
                self.load_swc()
                self.save_cache()
        else:
            # Load SWC file
            self.write_log("Ignoring old cache, rewriting new cache file")
            self.load_swc()
            self.save_cache()

        self.place()  # Updates position and rotation

        # Remove axonStumpIDFlag completely later...
        assert not self.axon_stump_id_flag, \
            "axonStumpFlag is depricated, should be off"

    ############################################################################

    def clone(self,
              load_morphology=None,  # True or False, None = same as parent
              position=np.zeros((1, 3)),
              rotation=None,
              morphology_id=None,
              parameter_id=None,
              modulation_id=None,
              parameter_key=None,
              morphology_key=None,
              modulation_key=None):

        """
        Creates a clone copy of a neuron.

        Args:
            load_morphology (bool) : Load morphology into clone?
            position (float,float,float) : x,y,z coordinate of clone
            rotation (rotation matrix) : Rotation matrix for clone

            morphology_id: Morphology ID for the clone
            parameter_id: Parameter ID for the clone
            modulation_id: Neuromodulation parameter ID for the clone

            parameter_key (str): Parameter Key for clone
            morphology_key (str): Morphology Key for clone
            modulation_key (str): Modulation Key for clone

        """

        if load_morphology is None:
            load_morphology = self.load_morphology

        # If these are explicitly set to None, reuse to original coordinates
        # and rotation
        if position is None:
            position = self.position

        if rotation is None:
            rotation = self.rotation

        new_neuron = NeuronMorphology(name=self.name,
                                      position=position,
                                      rotation=rotation,
                                      swc_filename=self.swc_filename,
                                      param_filename=self.param_filename,
                                      param_data=self.param_data,
                                      mech_filename=self.mech_filename,
                                      modulation=self.modulation,
                                      neuron_path=self.neuron_path,
                                      morphology_id=morphology_id,
                                      parameter_id=parameter_id,
                                      modulation_id=modulation_id,
                                      parameter_key=parameter_key,
                                      morphology_key=morphology_key,
                                      modulation_key=modulation_key,
                                      verbose=self.verbose,
                                      load_morphology=False,
                                      hoc=self.hoc,
                                      virtual_neuron=self.virtual_neuron)

        if load_morphology:
            # Set the flag
            new_neuron.load_morphology = load_morphology

            # Copy the data
            new_neuron.axon = np.copy(self.axon)
            new_neuron.dend = np.copy(self.dend)
            new_neuron.soma = np.copy(self.soma)

            # Warn the user if the neuron is already rotated
            new_neuron.rotated_flag = self.rotated_flag

            new_neuron.place()

            # These dont change either, so skip np.copy
            new_neuron.axon_links = self.axon_links
            new_neuron.dend_links = self.dend_links
            new_neuron.dend_sec_x = self.dend_sec_x
            new_neuron.dend_sec_id = self.dend_sec_id
            new_neuron.sec_id_links = self.sec_id_links
            new_neuron.sec_id_links_x = self.sec_id_links_x
            new_neuron.sec_id_to_len = self.sec_id_to_len

            new_neuron.axon_stump_id_flag = self.axon_stump_id_flag

        new_neuron.max_axon_radius = self.max_axon_radius
        new_neuron.max_dend_radius = self.max_dend_radius

        if self.dend_density is not None:
            new_neuron.dend_density = self.dend_density

        if self.axon_density is not None:
            new_neuron.axon_density = self.axon_density

        new_neuron.voxel_size = self.voxel_size

        if self.axon_density_type is not None:
            new_neuron.axon_density_type = self.axon_density_type

        if self.axon_density_bounds_xyz is not None:
            new_neuron.axon_density_bounds_xyz = self.axon_density_bounds_xyz

        return new_neuron

    ############################################################################

    def write_log(self, text, is_error=False):

        """ Write text to log file. Prints on screen if self.verbose or is_error """
        if self.log_file is not None:
            self.log_file.write(f"{text}\n")

        if self.verbose or is_error:
            print(text)

    ############################################################################

    # http://blog.lostinmyterminal.com/python/2015/05/12/random-rotation-matrix.html
    @staticmethod
    def rand_rotation_matrix(deflection=1.0, rand_nums=None):
        """
        Creates a random rotation matrix.
    
        deflection: the magnitude of the rotation. For 0, no rotation; for 1, competely random
        rotation. Small deflection => small perturbation.
        rand_nums: 3 random numbers in the range [0, 1]. If `None`, they will be auto-generated.
        """

        # from http://www.realtimerendering.com/resources/GraphicsGems/gemsiii/rand_rotation.c

        if rand_nums is None:
            rand_nums = np.random.uniform(size=(3,))

        theta, phi, z = rand_nums

        theta = theta * 2.0 * deflection * np.pi  # Rotation about the pole (Z).
        phi = phi * 2.0 * np.pi  # For direction of pole deflection.
        z = z * 2.0 * deflection  # For magnitude of pole deflection.

        # Compute a vector V used for distributing points over the sphere
        # via the reflection I - V Transpose(V).  This formulation of V
        # will guarantee that if x[1] and x[2] are uniformly distributed,
        # the reflected points will be uniform on the sphere.  Note that V
        # has length sqrt(2) to eliminate the 2 in the Householder matrix.

        r = np.sqrt(z)
        vv = (np.sin(phi) * r, np.cos(phi) * r, np.sqrt(2.0 - z))

        st = np.sin(theta)
        ct = np.cos(theta)

        rr = np.array(((ct, st, 0), (-st, ct, 0), (0, 0, 1)))

        # Construct the rotation matrix  ( V Transpose(V) - I ) R.
        mm = (np.outer(vv, vv) - np.eye(3)).dot(rr)
        return mm

    ############################################################################

    # We can specify a position and rotation
    def place(self, rotation=None, position=None):

        """
        Placing a neuron and rotating it. Will give a warning if it was previously rotated.

        Args:
            rotation: 3D rotation matrix
            position: x,y,z position for neuron

        """

        if self.rotated_flag:
            self.write_log("!!! WARNING, rotating a rotated neuron...")

        if rotation is None:
            rotation = self.rotation
        elif type(rotation) is not np.ndarray:
            rotation = np.array(rotation)

        if position is None:
            position = self.position
        elif type(position) is not np.ndarray:
            position = np.array(position)

        # print("Place called! pos = " + str(position) + ", rot = " + str(rotation))

        # rotation = self.randRotationMatrix()

        # We subtract soma before rotating to centre neuron
        if rotation is not None:

            assert np.abs(np.linalg.det(rotation) - 1) < 1e-6, \
                (f"place: determinant of rotation matrix should be 1 (did you miss matmul when multiplying?)"
                 f"rotation = {rotation}, det(rotation) = {np.linalg.det(rotation)}")

            self.rotated_flag = True

            if len(self.axon) > 0:
                self.axon[:, 0:3] = \
                    np.transpose(np.matmul(rotation, np.transpose(self.axon[:, 0:3] - self.soma[0, 0:3])))
            if len(self.dend) > 0:
                self.dend[:, 0:3] = \
                    np.transpose(np.matmul(rotation, np.transpose(self.dend[:, 0:3] - self.soma[0, 0:3])))
            if len(self.soma) > 0:
                self.soma[:, 0:3] = \
                    np.transpose(np.matmul(rotation, np.transpose(self.soma[:, 0:3] - self.soma[0, 0:3])))

        # Place neuron in correct position
        if len(self.axon) > 0:
            self.axon[:, 0:3] = self.axon[:, 0:3] - self.soma[0, 0:3] + position

        if len(self.dend) > 0:
            self.dend[:, 0:3] = self.dend[:, 0:3] - self.soma[0, 0:3] + position

        if len(self.soma) > 0:
            self.soma[:, 0:3] = self.soma[:, 0:3] - self.soma[0, 0:3] + position

        # Track rotation and location
        self.rotation = rotation
        self.position = position

        # Plot neuron post rotation
        # self.plot_neuron()

        return self

    ############################################################################

    def save_cache(self, cache_file=None):

        """ Saves cache_file with morphology """

        if cache_file is None:
            cache_file = snudda_parse_path(self.cache_filename, self.snudda_data)

        if cache_file is None:
            self.write_log("Unable to save neuron cache file, no cache_file name specified.")
            return

        assert not self.rotated_flag, \
            "saveCache: The neuron should not be rotated when saving cache"

        morph = dict([])

        morph["swc_filename"] = self.swc_filename
        morph["soma"] = self.soma
        morph["axon"] = self.axon
        morph["dend"] = self.dend
        morph["axonLinks"] = self.axon_links
        morph["dendLinks"] = self.dend_links
        morph["dendSecX"] = self.dend_sec_x
        morph["dendSecID"] = self.dend_sec_id
        morph["axonStumpIDFlag"] = self.axon_stump_id_flag
        morph["maxAxonRadius"] = self.max_axon_radius
        morph["maxDendRadius"] = self.max_dend_radius
        morph["dendDensity"] = self.dend_density
        morph["axonDensity"] = self.axon_density

        morph["secIdToSecLen"] = self.sec_id_to_len
        morph["secIdLinks"] = self.sec_id_links
        morph["secIdLinksX"] = self.sec_id_links_x

        morph["version"] = self.cache_version

        assert (cache_file != self.swc_filename)

        self.write_log(f"Saving cache file: {cache_file}")

        try:
            import pickle
            with open(cache_file, 'wb') as cache_file:
                pickle.dump(morph, cache_file, self.pickle_version)
        except:
            self.write_log(f"Unable to save cache file {cache_file} -- do you have permission to write?")

    ############################################################################

    def cache_exist(self, cache_file=None):

        """ Checks if cache_file exists """

        if cache_file is None:
            cache_file = snudda_parse_path(self.cache_filename, self.snudda_data)

        cache_flag = False

        import os

        if os.path.isfile(cache_file):

            swc_time = os.path.getmtime(snudda_parse_path(self.swc_filename, self.snudda_data))
            cache_time = os.path.getmtime(cache_file)

            if cache_time > swc_time:
                self.write_log(f"Found cache file: {cache_file}")
                cache_flag = True
            else:
                self.write_log(f"Found old cache file: {cache_file}")

        else:
            self.write_log("No cache file found.")

        return cache_flag

    ############################################################################

    def load_cache(self, cache_file=None):

        """ Loads morphology from cache_file. """

        if cache_file is None:
            cache_file = snudda_parse_path(self.cache_filename, self.snudda_data)

        assert cache_file is not None, "Unable to open cache file, cache file name not set."

        import pickle
        with open(cache_file, 'rb') as cache_file:
            morph = pickle.load(cache_file)

        assert self.swc_filename == morph["swc_filename"], \
            "Cached file had different path. Saving new version of cache."

        assert self.axon_stump_id_flag == morph["axonStumpIDFlag"], \
            "axonStumpIDFlag must match cached version"

        # axonStumpIDFlag affects the section ID for the dendrites (and axon)
        # True when running Network_simulate.py and False if running Neurodamus.

        # self.axonStumpIDFlag = morph["axonStumpIDFlag"] # True or False

        self.soma = np.copy(morph["soma"])
        self.axon = np.copy(morph["axon"])
        self.dend = np.copy(morph["dend"])

        self.axon_links = morph["axonLinks"]
        self.dend_links = morph["dendLinks"]
        self.dend_sec_x = morph["dendSecX"]
        self.dend_sec_id = morph["dendSecID"]

        assert morph["version"] == self.cache_version, \
            "Cache version mismatch, regenerating cache"

        self.max_axon_radius = morph["maxAxonRadius"]
        self.max_dend_radius = morph["maxDendRadius"]

        if morph["dendDensity"] is not None:
            self.dend_density = morph["dendDensity"]
        else:
            self.dend_density = None

        if morph["axonDensity"] is not None:
            self.axon_density = morph["axonDensity"]
        else:
            self.axon_density = None

        self.sec_id_to_len = morph["secIdToSecLen"]

        self.sec_id_links = morph["secIdLinks"]
        self.sec_id_links_x = morph["secIdLinksX"]

        # Place neuron -- Do not place neuron, loadNeuronMorphology does that
        # self.place()

    ############################################################################

    # self.actionStumpIDFlag only affects the section ID.
    # If it is set to False, all sectionID are computed normally
    # if it is set to True, each axon will have the same sectionID throughout
    # if there are multiple axons they will have separate sectionIDs

    def load_swc(self, swc_file=None):

        """ Loads morphology from swc_file """

        if not swc_file:
            swc_file = snudda_parse_path(self.swc_filename, self.snudda_data)

        with open(swc_file, 'r') as f:
            lines = f.readlines()

        comp_type = {1: "soma", 2: "axon", 3: "dend", 4: "apic"}

        swc_vals = np.zeros(shape=(len(lines), 7))

        num_comps = 0
        for ss in lines:
            if ss[0] != '#':
                swc_vals[num_comps, :] = [float(s) for s in ss.split()]
                num_comps = num_comps + 1

        # swcVals -- 0: compID, 1: type, 2,3,4: xyz coords, 5: radius, 6: parentID
        assert (1 <= swc_vals[:num_comps, 1]).all() and (swc_vals[:num_comps, 1] <= 4).all(), \
            f"loadMorphology: Only types 1,2,3,4 are supported: {swc_file}"

        # Subtract 1 from ID and parentID, so we get easier indexing
        swc_vals[:, 0] -= 1
        swc_vals[:, 6] -= 1

        swc_vals[:, 2:6] *= 1e-6  # Convert to meter x,y,z, radie

        # Columns:
        # 0: ID, 1,2,3: x,y,z 4: radie, 5: type, 6: parent, 7: somaDist,
        # 8: nodeParent, 9: childCount, 10: sectionID, 11: sectionLen,
        # 12: segmentLen

        # -- careful with sectionID and sectionLen at branch points,
        #    they belong to the parent section
        # -- also dont confuse sectionLen and segmentLen (the latter is
        #    for the segment, which is a part of the larger section)
        points = np.zeros((num_comps, 13))
        points[:num_comps, 0] = swc_vals[:num_comps, 0]  # ID
        points[:num_comps, 1:5] = swc_vals[:num_comps, 2:6]  # x,y,z,r
        points[:num_comps, 5] = swc_vals[:num_comps, 1]  # type
        points[:num_comps, 6] = swc_vals[:num_comps, 6]  # parent

        assert points[0, 5] == 1, \
            "First compartment must be a soma: " + str(swc_file)

        # Create list of the links,
        # exclude soma -> first comp link (should be within soma radius)
        # Columns: 0: ID1, 1: ID2, 2: sectionID, 3: sectionX0, 4: sectionX1
        # 5: nodeParent, 6:type

        links = np.zeros((num_comps, 7))

        link_idx = 0
        for idx in range(0, num_comps):
            id0 = int(points[idx, 6])  # parent
            id1 = int(points[idx, 0])  # point

            if id0 <= 0:
                # No parent or soma is parent, skip link
                continue

            links[link_idx, 0:2] = [id0, id1]
            links[link_idx, 5] = points[idx, 6]

            link_idx += 1

        # Trim link list
        links = links[:link_idx, :]

        # Count children each node has
        for idx in range(1, num_comps):
            try:
                # Increment parents child counter
                points[int(points[idx, 6]), 9] += 1
            except:
                self.write_log(f"Are there gaps in the numbering of the compartments in the SWC file: {swc_file}",
                               is_error=True)
                import traceback
                tstr = traceback.format_exc()
                self.write_log(tstr)
                import pdb
                pdb.set_trace()

        # Make sure soma has a child count > 1 --- no we dont want soma as node
        # if(points[0,9] == 0):
        #   points[0,9] = 100

        # Also make sure all points with soma as parent get child count > 1
        # (Child Count > 1 ==> start or end of segment)
        soma_child_idx = np.where(points[:, 6] == 0)[0]
        points[soma_child_idx, 9] += 50

        # Mark node parent, and assign sectionID
        # -- this is used to set sectionID for links, the link
        # will use the end points sectionID
        # !!! Make sure sectionID is correct, and match what Neuron uses internally

        # Nodes are branch points (> 1 child), or end points (0 children)
        # and not soma
        node_idx = np.where((points[:, 9] != 1) & (points[:, 5] != 1))[0]

        # soma is section 0, but we dont include connection soma to first node
        # so let the first dend node be 0, since the section ID is taken from
        # the child ID
        section_id = 1

        # Sonata specifies first axon, then basal, then apical sections
        axon_idx = node_idx[np.where(points[node_idx, 5] == 2)[0]]
        basal_idx = node_idx[np.where(points[node_idx, 5] == 3)[0]]
        apical_idx = node_idx[np.where(points[node_idx, 5] == 4)[0]]

        # If simulation will use an axon stump, where each axon branch is shortened
        # to a stump with the same section ID, then we need to make sure the
        # numbering is correct for the dendrites.

        # Update, set axonID to -1
        for nIdx in axon_idx:
            points[nIdx, 10] = -1

        # Set soma ID to 0
        points[0, 10] = 0

        # Calculate sectionID for dendrites
        section_id = 1

        # Axon dealt with, only loop over dendrites next
        node_loop_list = [basal_idx, apical_idx]

        for idxList in node_loop_list:
            for nIdx in idxList:
                if points[nIdx, 6] > 0:
                    # Set section ID, exclude soma, and compartments bordering to soma
                    points[nIdx, 10] = section_id
                    section_id += 1

        # Assign node parents
        for nIdx in node_idx:

            # Find node parent
            parent_idx = int(points[nIdx, 6])
            # While one child (= no node), keep following parent
            # But stop if parent is soma, or if grandparent is soma
            # !!! Here last link node to soma is not included in neurite morphology
            #     since we assume it is inside the soma
            while points[parent_idx, 9] == 1 and parent_idx > 0 and points[parent_idx, 6] > 0:
                parent_idx = int(points[parent_idx, 6])

            node_parent_idx = parent_idx
            points[nIdx, 8] = node_parent_idx

            section_id = points[nIdx, 10]
            parent_idx = int(points[nIdx, 6])
            while points[parent_idx, 9] == 1 and parent_idx > 0:
                points[parent_idx, 8] = node_parent_idx
                assert points[parent_idx, 10] == 0, "SectionID should be unset prior"
                points[parent_idx, 10] = section_id
                parent_idx = int(points[parent_idx, 6])

        for idx in range(1, num_comps):
            parent_idx = int(points[idx, 6])

            # Calculate soma dist (and also save segLen)
            seg_len = np.sqrt(np.sum((points[idx, 1:4] - points[parent_idx, 1:4]) ** 2))
            points[idx, 7] = points[parent_idx, 7] + seg_len
            points[idx, 12] = seg_len

        # Calculate section length (length between nodes)
        for idx in node_idx:
            node_parent_idx = int(points[idx, 8])

            # Difference in soma distance is section length
            section_len = points[idx, 7] - points[node_parent_idx, 7]
            points[idx, 11] = section_len

            if section_len == 0:
                self.write_log("Section length is zero --- !!! ")
                import pdb
                pdb.set_trace()

            prev_idx = int(points[idx, 6])
            while prev_idx > node_parent_idx:
                points[prev_idx, 11] = section_len
                prev_idx = int(points[prev_idx, 6])

        # Calculate sectionX
        for idx in range(0, links.shape[0]):
            id0 = int(links[idx, 0])
            id1 = int(links[idx, 1])
            links[idx, 2] = points[id1, 10]  # Section ID from point (not parent)

            node_parent = int(points[id1, 8])
            node_parent_soma_dist = points[node_parent, 7]
            section_len = points[id1, 11]

            # segX0 and segX1
            links[idx, 3] = (points[id0, 7] - node_parent_soma_dist) / section_len
            links[idx, 4] = (points[id1, 7] - node_parent_soma_dist) / section_len

            links[idx, 5] = node_parent
            links[idx, 6] = points[id0, 5]  # type (use parent,
            # to avoid soma to dend link)

        # Store the soma, axon, dend and links in the object

        self.soma = np.zeros((1, 4))
        self.soma[0, :] = swc_vals[0, 2:6]  # save x,y,z,r

        if np.linalg.norm(self.soma[0, :3]) > 0:
            self.write_log(f"WARNING {swc_file} has soma which is not centered at (0,0,0)", is_error=True)

        dend_idx = np.where((points[:, 5] == 3) | (points[:, 5] == 4))[0]
        axon_idx = np.where(points[:, 5] == 2)[0]

        dend_link_idx = np.where((links[:, 6] == 3) | (links[:, 6] == 4))[0]
        axon_link_idx = np.where(links[:, 6] == 2)[0]

        # 0,1,2: x,y,z  3: radie, 4: dist to soma
        self.dend = np.zeros((len(dend_idx), 5))
        self.axon = np.zeros((len(axon_idx), 5))

        self.dend_links = np.zeros((len(dend_link_idx), 2), dtype=int)  # ID0,ID1
        self.axon_links = np.zeros((len(axon_link_idx), 2), dtype=int)  # ID0,ID1

        self.dend_sec_id = np.zeros((len(dend_link_idx),), dtype=int)  # SectionID
        self.dend_sec_x = np.zeros((len(dend_link_idx), 2))  # SecX0, SecX1

        dend_lookup = dict([])
        axon_lookup = dict([])

        for idx in range(0, len(dend_idx)):
            dend_lookup[dend_idx[idx]] = idx

        for idx in range(0, len(axon_idx)):
            axon_lookup[axon_idx[idx]] = idx

        for idx, d_idx in enumerate(dend_idx):
            self.dend[idx, 0:4] = points[d_idx, 1:5]  # x,y,z,r
            self.dend[idx, 4] = points[d_idx, 7]  # dist to soma

        for idx, a_idx in enumerate(axon_idx):
            self.axon[idx, 0:4] = points[a_idx, 1:5]  # x,y,z,r
            self.axon[idx, 4] = points[a_idx, 7]  # dist to soma

        for idx, d_idx in enumerate(dend_link_idx):
            self.dend_links[idx, 0] = dend_lookup[int(links[d_idx, 0])]  # ID0 - parent
            self.dend_links[idx, 1] = dend_lookup[int(links[d_idx, 1])]  # ID1

            self.dend_sec_id[idx] = links[d_idx, 2]
            self.dend_sec_x[idx, :] = links[d_idx, 3:5]

        for idx, a_idx in enumerate(axon_link_idx):
            self.axon_links[idx, 0] = axon_lookup[links[a_idx, 0]]
            self.axon_links[idx, 1] = axon_lookup[links[a_idx, 1]]
            # We also have sectionID, secX0 and secX1 saved in links[:,2:5]
            # if needed in the future

        if self.virtual_neuron:
            # For virtual neurons, skip the dendrites (save space)
            self.dend = np.zeros((0, self.dend.shape[1]))
            self.dend_links = np.zeros((0, 2))
            self.dend_sec_id = np.zeros((0,))
            self.dend_sec_x = np.zeros((0, 2))

        # Lookup for section length based on section id
        self.sec_id_to_len = np.ones((int(1 + np.max(points[:, 10])),))
        for sec_id, sec_len in zip(points[:, 10].astype(int), points[:, 11]):
            if sec_id >= 0:
                self.sec_id_to_len[sec_id] = sec_len

        # Lookup table to find self.dend_links with a given section id
        tmp_id_lookup = dict()
        tmp_x_lookup = dict()
        for idx, (sec_id, sec_x) in enumerate(zip(self.dend_sec_id, self.dend_sec_x)):
            if sec_id not in tmp_id_lookup:
                tmp_id_lookup[sec_id] = [idx]
                tmp_x_lookup[sec_id] = [sec_x[1]]
            else:
                tmp_id_lookup[sec_id].append(idx)
                tmp_x_lookup[sec_id].append(sec_x[1])

        for k in tmp_id_lookup.keys():
            self.sec_id_links[k] = np.array(tmp_id_lookup[k])
            self.sec_id_links_x[k] = np.array(tmp_x_lookup[k])

        # self.dendriteDensity() # -- depricated
        self.find_radius()
        self.place()

        # self.debug_plot()

    ############################################################################

    def get_section_coordinates(self, section_id, section_x):

        if section_id == 0:
            return self.soma[0, 0:3]

        assert 0 <= section_x <= 1, f"section_x should be between 0, 1. section_x={section_x}"

        # Find the relevant dendrite link
        link_idx = np.where(np.logical_and(self.dend_sec_id == section_id,
                                           np.logical_and(self.dend_sec_x[:, 0] <= section_x,
                                                          section_x <= self.dend_sec_x[:, 1])))[0]

        assert len(link_idx) == 1, \
            (f"Unable to find a compartment matching section_id={section_id}, section_x={section_x}."
             f" Found {len(link_idx)}: {link_idx}."
             f"\nMorphology file: {self.swc_filename}")

        assert self.dend_sec_id[link_idx[0]] == section_id

        p1 = self.dend[self.dend_links[link_idx[0], 0], :3]
        p2 = self.dend[self.dend_links[link_idx[0], 1], :3]

        coords = p2 * section_x + (1 - section_x) * p1

        return coords

    ############################################################################

    def find_radius(self):

        """ Find finds maximum axon and dendrite radius of neuron. """

        if len(self.axon) > 0:
            self.max_axon_radius = \
                np.max(np.linalg.norm(self.axon[:, 0:3] - self.soma[0, 0:3], axis=1))

        if len(self.dend) > 0:
            self.max_dend_radius = \
                np.max(np.linalg.norm(self.dend[:, 0:3] - self.soma[0, 0:3], axis=1))
        else:
            self.max_dend_radius = 0

        if self.verbose:
            self.write_log(f"Max axon radius = {self.max_axon_radius}")
            self.write_log(f"Max dend radius = {self.max_dend_radius}")

    ############################################################################

    def plot_neuron(self,
                    axis=None,
                    plot_axon=True,
                    plot_dendrite=True,
                    line_style='-',
                    alpha=1.0,
                    plot_origo=None,  # Only use this when plotting hyper voxels
                    plot_scale=1.0,
                    axon_colour=None,
                    dend_colour=None,
                    soma_colour=None,
                    show_plot=True):

        """
        Plots neuron.

        Args:
            axis
            plot_axon
            plot_dendrite
            line_style
            alpha
            plot_origo
            plot_scale
            axon_colour
            dend_colour
            soma_colour
            show_plot
        """

        if plot_origo is None:
            plot_origo = np.array([0, 0, 0])

        self.write_log(f"Plotting neuron {self.swc_filename}")

        if axon_colour is None:
            axon_colour = self.colour
        if dend_colour is None:
            dend_colour = self.colour
        if soma_colour is None:
            soma_colour = self.colour

        import matplotlib.pyplot as plt

        if axis is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
        else:
            ax = axis

        if len(self.axon) > 0 and plot_axon:
            ax_links = []

            for row in self.axon_links[:, :2].astype(int):
                if len(ax_links) == 0:
                    ax_links = list(row)
                elif row[0] == ax_links[-1]:
                    ax_links.append(row[1])
                elif row[1] == ax_links[-1]:
                    ax_links.append(row[0])
                else:
                    ax.plot((self.axon[ax_links, 0] - plot_origo[0]) * plot_scale,
                            (self.axon[ax_links, 1] - plot_origo[1]) * plot_scale,
                            (self.axon[ax_links, 2] - plot_origo[2]) * plot_scale,
                            linestyle=line_style,
                            marker=',',
                            alpha=alpha,
                            c=axon_colour)

                    ax_links = list(row)

            if len(ax_links) > 0:
                ax.plot((self.axon[ax_links, 0] - plot_origo[0]) * plot_scale,
                        (self.axon[ax_links, 1] - plot_origo[1]) * plot_scale,
                        (self.axon[ax_links, 2] - plot_origo[2]) * plot_scale,
                        linestyle=line_style,
                        marker=',',
                        alpha=alpha,
                        c=axon_colour)

                # TODO: Also connect first axon compartment to soma

        if plot_dendrite:
            dend_links = []
            for row in self.dend_links[:, :2].astype(int):
                if len(dend_links) == 0:
                    dend_links = list(row)
                elif row[0] == dend_links[-1]:
                    dend_links.append(row[1])
                elif row[1] == dend_links[-1]:
                    dend_links.append(row[0])
                else:
                    ax.plot((self.dend[dend_links, 0] - plot_origo[0]) * plot_scale,
                            (self.dend[dend_links, 1] - plot_origo[1]) * plot_scale,
                            (self.dend[dend_links, 2] - plot_origo[2]) * plot_scale,
                            linestyle=line_style,
                            marker=',',
                            alpha=alpha,
                            c=dend_colour)

                    dend_links = list(row)

            if len(dend_links) > 0:
                ax.plot((self.dend[dend_links, 0] - plot_origo[0]) * plot_scale,
                        (self.dend[dend_links, 1] - plot_origo[1]) * plot_scale,
                        (self.dend[dend_links, 2] - plot_origo[2]) * plot_scale,
                        linestyle=line_style,
                        marker=',',
                        alpha=alpha,
                        c=dend_colour)

                # TODO: Also connect dendrites to soma for plot

        if len(self.soma) > 0:

            if self.soma.shape[0] == 1:
                u, v = np.mgrid[0:2 * np.pi:20j, 0:np.pi:10j]
                x = (self.soma[0, 3] * np.cos(u) * np.sin(v) + self.soma[0, 0] - plot_origo[0]) * plot_scale
                y = (self.soma[0, 3] * np.sin(u) * np.sin(v) + self.soma[0, 1] - plot_origo[1]) * plot_scale
                z = (self.soma[0, 3] * np.cos(v) + self.soma[0, 2] - plot_origo[2]) * plot_scale

                ax.plot_wireframe(x, y, z, color=soma_colour, alpha=alpha)
            else:
                ax.scatter((self.soma[:, 0] - plot_origo[0]) * plot_scale,
                           (self.soma[:, 1] - plot_origo[1]) * plot_scale,
                           (self.soma[:, 2] - plot_origo[2]) * plot_scale,
                           c=soma_colour, alpha=alpha)

        if axis is None:
            plt.title("Neuron: " + self.swc_filename.split("/")[-3] + "_" + self.swc_filename.split('/').pop())

            if show_plot:
                plt.ion()
                plt.show()
                plt.draw()
                plt.pause(0.001)

        return ax

    ############################################################################

    def set_axon_voxel_radial_density(self, density, max_axon_radius):

        """
        Sets axon radial density

        Args:
            density: Axon density f(r), r = radius from soma
            max_axon_radius: Axon density is calculated within a sphere of radius max_axon_radius

        """

        self.write_log("Only saving equation now")

        self.axon_density_type = "r"
        self.axon_density = density
        self.max_axon_radius = max_axon_radius

    ############################################################################

    def set_axon_voxel_xyz_density(self,
                                   density,
                                   axon_density_bounds_xyz):

        """
        Sets axon density

        Args:
            density: Axon density f(x,y,z), x,y,z = SWC coordinates in relative to soma
            axon_density_bounds_xyz: Bounding box for the axon density in x,y,z
        """

        self.write_log("Only saving equation now")

        self.axon_density_type = "xyz"
        self.axon_density = density
        self.axon_density_bounds_xyz = axon_density_bounds_xyz

    ############################################################################

    def compartment_length(self, comp_type="dend"):

        """
        Calculates compartment length comp_type ("axon" or "dend")
        """

        if comp_type == "dend":
            links = self.dend_links
            coords = self.dend
        elif comp_type == "axon":
            links = self.axon_links
            coords = self.axon
        else:
            assert False, "Unknown compartment type: " + str(comp_type) \
                          + ", valid types are 'axon' and 'dend'"

        comp_len = np.linalg.norm(coords[links[:, 0], :][:, :3] - coords[links[:, 1], :][:, :3], axis=1)

        return comp_len

    ############################################################################

    # TODO: Update the code so that it gives exactly num_locations positions (currently it varies)

    def dendrite_input_locations(self, synapse_density, rng, num_locations=None,
                                 cluster_size=None, cluster_spread=20e-6):

        """
        Randomises input locations on dendrites.

        Args:
            synapse_density : Synapse density as a function f(d), d=distance from soma
            rng : Numpy random stream
            num_locations : Number of input locations (this is average number returned, results vary)
            cluster_size (int): Number of synapse clusters to place (None = no clusters, all placed independently)
        """

        if num_locations is not None:
            # This function returns the exact number of synapses specified
            return self.dendrite_input_locations_helper(synapse_density=synapse_density,
                                                        rng=rng,
                                                        num_locations=num_locations,
                                                        cluster_size=cluster_size,
                                                        cluster_spread=cluster_spread)

        expected_synapses = self.get_expected_synapses_per_compartment(synapse_density=synapse_density)

        if num_locations is not None:
            expected_synapses *= num_locations / np.sum(expected_synapses)

        if cluster_size is not None:
            expected_synapses /= cluster_size
        else:
            cluster_size = 1

        # Number of input synapses on each compartment
        number_of_synapses = (expected_synapses + ((expected_synapses % 1)
                                                   > rng.random(len(expected_synapses)))).astype(int)

        n_syn_tot = np.sum(number_of_synapses)
        dist_syn_soma = []

        # x,y,z, secID, secX
        input_loc = np.zeros((n_syn_tot * cluster_size, 5))
        d = self.dend[:, 4]

        # Iterate over each compartment
        syn_ctr = 0
        for i_comp, n_syn in enumerate(number_of_synapses):

            # Add synapses to that compartment
            for j in range(0, n_syn * cluster_size):
                # print('Compartment containing a synapse',iComp)
                # print('Distance from soma',self.dend[iComp][4]*1e6,'$mum$')
                input_loc[syn_ctr, 3] = self.dend_sec_id[i_comp]

                # Cant have at endpoints 0 or 1
                comp_x = rng.random()
                dist_syn_soma = np.append(dist_syn_soma,
                                          d[self.dend_links[i_comp, 0]] * (1 - comp_x)
                                          + d[self.dend_links[i_comp, 1]] * comp_x)

                coords = (self.dend[self.dend_links[i_comp, 0], :3] * (1 - comp_x)
                          + self.dend[self.dend_links[i_comp, 1], :3] * comp_x)

                input_loc[syn_ctr, :3] = coords

                # Use compX (where between comp endpoints) to calculate sectionX
                # (where between section end points)
                input_loc[syn_ctr, 4] = self.dend_sec_x[i_comp, 0] * (1 - comp_x) + comp_x * self.dend_sec_x[i_comp, 1]

                syn_ctr += 1

        assert syn_ctr == input_loc.shape[0], f"Not all input_loc was set. Rows {input_loc.shape[0]}, syn_ctr={syn_ctr}"

        # if return_density:
        #     # Return xyz,secID,secX,iDensity,distSynSoma
        #     return input_loc[:, :3], input_loc[:, 3], input_loc[:, 4], i_density, dist_syn_soma

        # Return xyz,secID,secX, dist_to_soma (update: now also added distance synapse to soma)
        return input_loc[:, :3], input_loc[:, 3], input_loc[:, 4], dist_syn_soma

    ############################################################################

    def get_expected_synapses_per_compartment(self, synapse_density):

        # Calculate the input density at each point in dendrite morphology
        d = self.dend[:, 4]
        try:
            # d is now distance from some, so synapseDensity is a func of d
            i_density = numexpr.evaluate(synapse_density)
        except:
            self.write_log(f"Bad synapse density string: {synapse_density}")
            import traceback
            tstr = traceback.format_exc()
            self.write_log(tstr)
            assert False, f"Problem with synapse density {synapse_density}"

        # if type(i_density) in (int, float): -- this worked for eval, but not for numexpr.evaluate
        if i_density.ndim == 0:
            # If iDensity is a constant, we need to set it for all points
            i_density = float(i_density) * np.ones(d.shape)    # TODO: Change to np.full(d.shape, i_density, dtype=float)

        comp_density = (i_density[self.dend_links[:, 0]] + i_density[self.dend_links[:, 1]]) / 2
        comp_len = self.compartment_length(comp_type="dend")

        # comp_density is in synapses per micrometer, multiply by 1e6
        expected_synapses = comp_density * comp_len * 1e6

        return expected_synapses

    ############################################################################

    def cluster_synapses(self, sec_id, sec_x, count, distance, rng):

        """
            Randomize sec_x for cluster of 'count' synapses with centre placed at 'sec_id', 'sec_x'
            spread over 'distance' (but constrained to current section extent).

            Args:
                sec_id : Section id of cluster centre
                sec_x : Section x of cluster centre
                count : Number of synapses in cluster
                distance : Maximal spread of cluster along dendrite
                rng : Numpy random stream

            Returns:
                cluster_sec_x : Section x for cluster synapse
                coords : Coordinates for synapse (in meters)
                soma_dist : Distance to soma (in meters)
        """

        sec_len = self.sec_id_to_len[sec_id]
        min_sec_x = np.maximum(1e-3, sec_x - 0.5 * distance / sec_len)
        max_sec_x = np.minimum(1 - 1e-3, sec_x + 0.5 * distance / sec_len)

        cluster_sec_x = rng.uniform(low=min_sec_x, high=max_sec_x, size=count)

        syn_coords = np.zeros((count, 3))
        soma_dist = np.zeros((count,))

        for i, sx in enumerate(cluster_sec_x):
            if sec_id == 0:
                syn_coords[i, :] = self.soma[0, :3]
                soma_dist[i] = 0  # We are on the soma
            else:
                # We also need to calculate the x,y,z coordinates and distance to soma
                idx = len(self.sec_id_links_x[sec_id]) - np.sum(sx < self.sec_id_links_x[sec_id])
                link_idx = self.sec_id_links[sec_id][idx]
                assert self.dend_sec_id[link_idx] == sec_id

                x_start, x_end = self.dend_sec_x[link_idx]
                comp_x = (x_start - sx) / (x_start - x_end)
                start_info = self.dend[self.dend_links[link_idx, 0], :]  # x,y,z,radie,soma dist
                end_info = self.dend[self.dend_links[link_idx, 1], :]
                syn_info = start_info * (1 - comp_x) + end_info * comp_x

                syn_coords[i, :] = syn_info[:3]
                soma_dist[i] = syn_info[4]

        # OBS, syn_coords in meters, and soma dist in meters also
        return cluster_sec_x, syn_coords, soma_dist

    ############################################################################

    def dendrite_input_locations_helper(self,
                                        synapse_density,
                                        rng,
                                        num_locations,
                                        cluster_size=1,
                                        cluster_spread=20e-6):

        """
        Helper function. Places num_location clusters, each containing cluster_size synapses. Density is relative, scaled.

        Args:
            synapse_density : Synapse density as a function f(d), d = distance on dendrite from soma
            rng : Numpy random stream
            num_locations (int) : Number of locations
            cluster_size (int) : Size of synapse clusters
        """

        expected_synapses = self.get_expected_synapses_per_compartment(synapse_density=synapse_density)

        if cluster_size is None:
            cluster_size = 1

        num_synapses = num_locations * cluster_size
        input_loc = np.zeros((num_synapses, 6))

        d = self.dend[:, 4]
        syn_ctr = 0

        assert np.all(expected_synapses >= 0), f"Check your synapse density {synapse_density}, found negative values."
        exp_sum = np.sum(expected_synapses)
        assert exp_sum > 0, f"Check your synapse density {synapse_density}, all elements zero."

        syn_idx = rng.choice(len(expected_synapses), size=num_locations, replace=True,
                             p=expected_synapses / exp_sum)

        for comp_idx in syn_idx:

            # Cluster centre sec_id:
            sec_id = self.dend_sec_id[comp_idx]
            # We place cluster centre in the middle of the 5 micrometer morphology piece
            sec_x = (self.dend_sec_x[comp_idx, 0] + self.dend_sec_x[comp_idx, 1]) / 2

            if cluster_size is None or cluster_size == 1:

                comp_x = rng.random()

                coords = (self.dend[self.dend_links[comp_idx, 0], :3] * (1 - comp_x)
                          + self.dend[self.dend_links[comp_idx, 1], :3] * comp_x)

                input_loc[syn_ctr, :3] = coords
                input_loc[syn_ctr, 3] = self.dend_sec_id[comp_idx]
                input_loc[syn_ctr, 4] = (self.dend_sec_x[comp_idx, 0] * (1 - comp_x)
                                         + comp_x * self.dend_sec_x[comp_idx, 1])
                input_loc[syn_ctr, 5] = (d[self.dend_links[comp_idx, 0]] * (1 - comp_x)
                                         + d[self.dend_links[comp_idx, 1]] * comp_x)

                syn_ctr += 1

            else:
                syn_info = self.cluster_synapses(sec_id=sec_id, sec_x=sec_x, count=cluster_size,
                                                 distance=cluster_spread, rng=rng)
                input_loc[syn_ctr:syn_ctr + cluster_size, :3] = syn_info[
                    1]  # Synapse x,y,z -- in global coordinate system(!)
                input_loc[syn_ctr:syn_ctr + cluster_size, 3] = sec_id  # Sec id
                input_loc[syn_ctr:syn_ctr + cluster_size, 4] = syn_info[0]  # Sec x
                input_loc[syn_ctr:syn_ctr + cluster_size, 5] = syn_info[2]  # Soma distance

                syn_ctr += cluster_size

        assert syn_ctr == input_loc.shape[0], f"Not all input_loc was set. Rows {input_loc.shape[0]}, syn_ctr={syn_ctr}"

        # Return xyz,secID,secX, dist_to_soma (update: now also added distance synapse to soma)
        return input_loc[:, :3], input_loc[:, 3], input_loc[:, 4], input_loc[:, 5]

    ############################################################################

    def place_cluster_helper(self, cluster_xyz, cluster_radius, cluster_size):

        raise NotImplementedError("This is not yet implemented")

        # 1. Identify points on neuron within the cluster radius (length)
        # 2. Pick n synapses from the positions available, restrict choice to segment id of
        #    parent/child/grandparent/grandchild segments
        # help: lookup matrix --> 1 x 5

    ############################################################################

    def debug_plot(self, wait_flag=False, plot_step=1, plot_axon_flag=False):

        ax = self.plot_neuron(plot_axon=plot_axon_flag, show_plot=False)
        import matplotlib.pyplot as plt

        if plot_axon_flag:
            for a in self.axon_links:
                x0 = self.axon[int(a[0]), 0:3]
                x1 = self.axon[int(a[1]), 0:3]
                x = (x0 + x1) / 2

                # ax.text(x=x0[0],y=x0[1],z=x0[2],s=str(np.around(a[3],2)),color='blue')
                # ax.text(x=x1[0],y=x1[1],z=x1[2],s=str(np.around(a[4],2)),color='red')
                ax.text(x=x[0], y=x[1], z=x[2], s=str(a[2]), color='black')

                self.write_log(f"ID: {a[2]}")
                # input(" ")

        ctr = 0
        for (d, dID, dX) in zip(self.dend_links, self.dend_sec_id, self.dend_sec_x):

            x0 = self.dend[int(d[0]), 0:3]
            x1 = self.dend[int(d[1]), 0:3]
            x = (x0 + x1) / 2

            # ax.text(x=x0[0],y=x0[1],z=x0[2],s=str(np.around(dX[0],2)),color='blue')
            # ax.text(x=x1[0],y=x1[1],z=x1[2],s=str(np.around(dX[1],2)),color='red')

            if ctr % plot_step == 0:
                ax.text(x=x[0], y=x[1], z=x[2], s=f"{dID}:{np.mean(dX):.2f}", color='black')
            ctr += 1

            self.write_log(f"ID: {dID} X = {np.around(dX[0], 2)} - {np.around(dX[1], 2)}")

            if wait_flag:
                input(" ")

        plt.show()

        return ax

    ############################################################################


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("file_name", help="Path to neuron file")
    parser.add_argument("--step", default=None,
                        help="Display segment ID and segment X for sections with segment step for text output (e.g. 10)",
                        type=int)
    args = parser.parse_args()

    nm = NeuronMorphology(swc_filename=args.file_name, verbose=True, use_cache=False)
    nm.place(rotation=nm.rand_rotation_matrix(), position=np.array([0, 0, 0]))

    if args.step is not None:
        nm.debug_plot(plot_step=args.step)
    else:
        import matplotlib.pyplot as plt

        nm.plot_neuron(show_plot=False)
        plt.show()




import os

import h5py

from snudda.utils.load import SnuddaLoad


class InspectInput(object):

    def __init__(self, network_file, input_file):

        self.network_file = network_file
        self.input_file = input_file

        # We just need this to identify which neuron is which
        self.network = SnuddaLoad(self.network_file, load_synapses=False)

        self.input_data = h5py.File(input_file, 'r')

    def get_morphologies(self):

        return [os.path.basename(c["morphology"]) for c in self.network.data["neurons"]]

    def get_input_types(self, cell_id):

        s_cell_id = [str(c) for c in cell_id]
        input_types = set()

        for sc_id in s_cell_id:
            inp_t = set([inp for inp in self.input_data["input"][sc_id]])
            input_types = input_types.union(inp_t)

        return list(input_types)

    def check_input_ratio(self, neuron_type, verbose=True):

        print(f"Counting inputs for {neuron_type}")

        cell_id = self.network.get_neuron_id_of_type(neuron_type)
        cell_id_str = [str(c) for c in cell_id]

        input_count = dict()

        cell_morph = self.get_morphologies()

        unique_morph = set([cell_morph[c] for c in cell_id])

        input_type_list = self.get_input_types(cell_id)

        for inp in input_type_list:
            input_count[inp] = dict()
            for um in unique_morph:
                input_count[inp][um] = 0

        morph_counter = dict()

        for um in unique_morph:
            morph_counter[um] = 0

        # !!! TODO: We should split this by morphology also...

        for c_id in self.input_data['input']:
            if c_id in cell_id_str:
                morph = cell_morph[int(c_id)]
                morph_counter[morph] += 1

                for inp in input_type_list:
                    if inp in self.input_data['input'][c_id]:
                        n_input = len(self.input_data['input'][c_id][inp]['nSpikes'])
                        input_count[inp][morph] += n_input

        for inp in input_type_list:
            for um in unique_morph:
                avg_inp = round(input_count[inp][um] / morph_counter[um], 1)
                print(f"{inp} morphology {um}: {avg_inp} inputs")

        return input_count


if __name__ == '__main__':
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Inspect input")
    parser.add_argument("networkFile", help="Network file (hdf5)", type=str)
    parser.add_argument("inputFile", help="Input file (hdf5)", type=str)
    parser.add_argument("neuronType", help="Neuron type", type=str)

    args = parser.parse_args()

    inspector = InspectInput(network_file=args.networkFile,
                             input_file=args.inputFile)

    inspector.check_input_ratio(neuron_type=args.neuronType)



import json
import os


class NeuromodulationSynapse:

    def __init__(self):
        self.synapse_modulation = dict()
        self.weight = None
        self.type = 'adaptive'

    def set_weight(self, weight):
        self.weight = weight

    def set_connection_type(self, neuromodulation_key, connector):
        """
                neurotransmitter_key is equivalent to the parameter used in the mod files, which marks the
                level, mod and maxMod parameters e.g. neurotransmitter_key, ACh, would have parameters levelACh,
                modACh and maxModACh in modulated mod files.
        """

        if connector in self.synapse_modulation.keys():
            raise KeyError('connector is already defined')

        self.synapse_modulation.update({neuromodulation_key: {'connector': connector,
                                                              'cells': dict()}})

    def add_cell_modulation(self, neuromodulation_key, cell, ion_channels=None, receptors=None, extrinsic=None,
                            type_connection=None):

        if ion_channels is None:
            ion_channels = dict(soma=list(), dendrite=list(), axon=list())

        if receptors is None:
            receptors = dict()

        if extrinsic is None:
            extrinsic = dict()

        self.synapse_modulation[neuromodulation_key]['cells'].update({cell: {
            'ion_channels': ion_channels,
            'receptors': receptors,
            'extrinsic': extrinsic,
            'type': type_connection}})

    def save(self, dir_path, name):

        cell_types = list()

        for neurotransmitter_type, modulation_information in self.synapse_modulation.items():

            for c in modulation_information["cells"]:
                if c not in cell_types:
                    cell_types.append(c)

        for neurotransmitter_type, modulation_information in self.synapse_modulation.items():
            for c in cell_types:
                if c not in modulation_information["cells"]:
                    modulation_information["cells"].update({c: {"ion_channels": dict(soma=list(), dendrite=list(), axon=list()),
                                                                "receptors": None,
                                                                "extrinsic": None,
                                                                "type": None}})
        temp = dict()
        temp.update({'type': self.type})
        temp.update({'description': self.synapse_modulation})
        temp.update({'weight': self.weight})
        with open(os.path.join(dir_path, name), 'w') as out_file:
            json.dump(temp, out_file, indent=4)


if __name__ == "__main__":
    sw = NeuromodulationSynapse()
    sw.set_weight(weight=0)

    # Acetylcholine

    sw.set_connection_type(connector="concACh", neuromodulation_key="ACh")

    sw.add_cell_modulation(neuromodulation_key="ACh",
                           cell="dSPN",
                           ion_channels={
                               "soma": ["kir_ms", "cal12_ms", "cal13_ms", "can_ms", "Im_ms"],
                               "dendrite": ["kir_ms", "cal12_ms", "cal13_ms"]},
                           type_connection="spiking-concentration")

    sw.add_cell_modulation(neuromodulation_key="ACh",
                           cell="iSPN",
                           ion_channels={
                               "soma": ["kir_ms", "cal12_ms", "cal13_ms", "can_ms"],
                               "dendrite": ["kir_ms", "cal12_ms", "cal13_ms"]},
                           type_connection="spiking-concentration")

    # Dopamine

    sw.set_connection_type(connector="concDA", neuromodulation_key="DA")

    sw.add_cell_modulation(neuromodulation_key="DA",
                           cell="dSPN",
                           ion_channels={
                               "soma": ["kas_ms", "kaf_ms", "can_ms"],
                               "dendrite": ["kaf_ms", "kas_ms"],
                               "axon": []},
                           receptors={"tmGabaA": {"maxMod": 0.8},
                                      "tmGlut": {"maxMod_AMPA": 1.2,
                                                 "maxMod_NMDA": 1.3,
                                                 "failRate": 0.7}},
                           extrinsic=["CorticalBase", "CorticalSignal", "Thalamic"],
                           type_connection="spiking-concentration")

    sw.save(dir_path="", name="DA-ACh-control.json")



import os
import numpy as np


def get_spn_fs_freq():

    data_dir = os.path.dirname(__file__)
    data_file = os.path.join(data_dir, "Kim et al., 2019 - Henry Yin - Extended2a.csv")

    data = np.genfromtxt(data_file, delimiter=",", filling_values=np.nan)
    spn_freq, fs_freq = data[:, 0], data[:, 1]
    fs_freq = fs_freq[~np.isnan(fs_freq)]

    return spn_freq, fs_freq


def resample_spn_freq(size, rng=None):

    spn_freq, _ = get_spn_fs_freq()

    if rng is None:
        rng = np.random.default_rng()

    resampled_freq = rng.choice(a=spn_freq, size=size, replace=True)

    return resampled_freq


def resample_fs_freq(size, rng=None):

    _, fs_freq = get_spn_fs_freq()

    if rng is None:
        rng = np.random.default_rng()

    resampled_freq = rng.choice(a=fs_freq, size=size, replace=True)

    return resampled_freq


def plot_freq_hist(data, title=None, bins=None):

    import matplotlib.pyplot as plt

    plt.figure()
    plt.hist(data, bins=bins)
    plt.xlabel("Frequency (Hz)")
    plt.ylabel("Frequency of occurence")
    if title:
        plt.title(title)

    plt.ion()
    plt.show()
    plt.pause(0.1)


if __name__ == "__main__":

    spn_freq, _ = get_spn_fs_freq()
    resampled_spn_freq = resample_spn_freq((10000,1))

    plot_freq_hist(spn_freq, "SPN frequency", bins=100)
    plot_freq_hist(resampled_spn_freq, "Resampled SPN frequency", bins=100)

    input("Press a key to continue")




import numpy as np
from matplotlib import pyplot as plt
import sys
import os
import json
from collections import OrderedDict

from snudda.plotting.plot_node_benchmark import PlotNodeBenchmark


class PlotSizeBenchmark(PlotNodeBenchmark):

    def __init__(self, network_path_list=None, network_size=None, size_prefix=True):

        assert len(network_path_list) == len(network_size)

        self.data = self.merge_data(map(self.load_benchmark, network_path_list), sort=False)
        self.network_size = np.array(network_size)
        self.fig_dir = os.path.join(network_path_list[0], "figures")
        self.size_prefix = size_prefix

        assert (self.network_size == np.sort(self.network_size)).all(), "Networks must be sorted by size"

    def plot_data(self):

        stage = ["Place", "Detect", "Prune"]

        nodes = self.data["place"][:, 1]
        assert (nodes == nodes[0]).all(), "All sizes should have same number of nodes"

        duration = np.zeros((len(stage), len(nodes)))

        for idx, s in enumerate(stage):
            duration[idx, :] = self.data[s.lower()][:, 0] / 3600
            assert (self.data[s.lower()][:, 1] == nodes).all(), f"Stage {s} is missing some node values that Place has"

        small_size = 15
        medium_size = 22
        bigger_size = 30

        plt.rc('font', size=small_size)  # controls default text sizes
        plt.rc('axes', titlesize=small_size)  # fontsize of the axes title
        plt.rc('axes', labelsize=medium_size)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=small_size)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=small_size)  # fontsize of the tick labels
        plt.rc('legend', fontsize=small_size)  # legend fontsize
        plt.rc('figure', titlesize=bigger_size)  # fontsize of the figure title

        fig = plt.figure()
        ax = plt.subplot()

        ax.stackplot(self.network_size, duration, labels=stage)
        ax.legend(loc="upper left")
        ax.set_xlabel("Network size (log scale)")
        ax.set_ylabel("Runtime (h)")
        ax.set_xscale("log")
        # ax.set_yscale("log")
        ax.set_xticks(self.network_size, minor=False)

        if self.size_prefix:
            ax.set_xticklabels(self.convert_labels(self.network_size))
        else:
            ax.set_xticklabels(self.network_size.astype(int))

        # Prevent cropping of x-label
        fig.subplots_adjust(bottom=0.15)

        if not os.path.isdir(self.fig_dir):
            os.mkdir(self.fig_dir)

        fig_name = os.path.join(self.fig_dir, "benchmark-size.pdf")
        plt.savefig(fig_name, dpi=300)
        plt.show()

    @staticmethod
    def convert_labels(values):

        prefix_lookup = [(1000000, "M"), (1000, "k")]
        converted = []

        for v in values:
            cv = f"{v}"
            for pval, prefix in prefix_lookup:
                if v % pval == 0:
                    cv = f"{int(v/pval)}{prefix}"
                    break
            converted.append(cv)

        return converted


if __name__ == "__main__":

    if len(sys.argv) > 2:
        network_path_list = sys.argv[1::2]
        network_size_list = [int(x) for x in sys.argv[2::2]]

        pb = PlotSizeBenchmark(network_path_list, network_size_list)
        pb.plot_data()

    else:
        print(f"Usage python3 {sys.argv[0]} network_path1 size1 network_path2 size2 ...")



import os
import numpy as np
import matplotlib.pyplot as plt

from snudda.plotting.plot_spike_raster_v2 import SnuddaPlotSpikeRaster2


class PlotPeriodExperiment(SnuddaPlotSpikeRaster2):

    def calculate_period_histogram_mod(self, period, neuron_id, time_range,
                                       exclude_depolarisation_blocked_neurons=False,
                                       n_bins=10):

        if exclude_depolarisation_blocked_neurons:
            if neuron_id is None:
                neuron_id = list(self.snudda_simulation_load.network_simulation_file["neurons"].keys())

            if self.snudda_simulation_load.depolarisation_block is None:
                self.snudda_simulation_load.depolarisation_block = self.snudda_simulation_load.check_depolarisation_block()

            bad_cells = sorted(list(set([x for x, ts, te in self.snudda_simulation_load.depolarisation_block])))
            neuron_id = np.array([x for x in neuron_id if x not in bad_cells])

        spikes = self.snudda_simulation_load.get_spikes(neuron_id=neuron_id)
        all_spikes = []

        n_spike_trains = len(spikes)

        for s in spikes.values():

            sf = s.flatten()

            if time_range is not None:
                idx = np.where(np.logical_and(time_range[0] <= sf, sf <= time_range[1]))[0]
                all_spikes = all_spikes + list(sf[idx] % period)
            else:
                all_spikes = all_spikes + list(sf % period)

        counts, bins = np.histogram(all_spikes, bins=n_bins)

        if time_range:
            t = time_range[1] - time_range[0]
        else:
            t = self.time[-1] - self.time[0]

        freq = counts/(n_spike_trains * (bins[1] - bins[0]) * t/period)

        return freq, bins

    def plot_period_histogram_mod(self, period, neuron_id=None, time_range=None, n_bins=20,
                                  fig_file=None, ax=None, fig_size=None, label=None, color=None,
                                  show_figure=True, exclude_depolarisation_blocked_neurons=False,
                                  save_figure=True, linestyle="-", legend_loc="best"):

        self.make_figures_directory()

        plt.rcParams.update({'font.size': 24,
                             'xtick.labelsize': 20,
                             'ytick.labelsize': 20,
                             'legend.loc': legend_loc})

        if ax is None:
            fig = plt.figure(figsize=fig_size)
            ax = fig.add_subplot()

        freq, bins = self.calculate_period_histogram_mod(period=period, neuron_id=neuron_id, time_range=time_range,
                                                         n_bins=n_bins,
                                                         exclude_depolarisation_blocked_neurons=exclude_depolarisation_blocked_neurons)

        ax.stairs(freq, bins * 1e3, label=label, color=color, linewidth=3, linestyle=linestyle)

        ax.set_xlabel("Time (ms)")
        ax.set_ylabel("Frequency (Hz)")
        ax.legend()

        if fig_file is None:
            fig_file = os.path.join(self.figure_path, "spike-period-histogram.pdf")
        else:
            fig_file = os.path.join(self.figure_path, fig_file)

        plt.tight_layout()

        if save_figure:
            print(f"Writing figure to {fig_file}")
            plt.savefig(fig_file, dpi=300)

        if show_figure:
            plt.ion()
            plt.show()

        # ax
        return ax, freq, bins

    def plot_period_voltage(self, period, neuron_type, ax=None, fig_size=None, figure_name=None, show_plot=True,
                            time_range=None, show_pre_histogram=True):

        """ This plots the average voltage for the neurons during the spikes, split into two groups. Those that
            had a presynaptic neuron spike before, and those that did not. """

        if ax is None:
            fig = plt.figure(figsize=fig_size)
            ax = fig.add_subplot()

        post_id_list = self.snudda_load.get_neuron_id_of_type(neuron_type=neuron_type)

        pre_id_lookup = dict()

        spike_with_pre_spike = []
        spike_no_pre_spike = []
        no_spike_with_pre_spike = []
        no_spike_no_pre_spike = []

        time = np.around(self.snudda_simulation_load.get_time(), 8)
        period_start = np.around(np.arange(0, time[-1], period), 8)  # Use around to avoid rounding errors
        period_end = np.around(period_start + period, 8)

        n_periods = int(np.ceil(time[-1] / period))
        spike_in_period = np.zeros((self.snudda_load.data["num_neurons"], n_periods))

        for neuron_id in self.snudda_load.get_neuron_id():
            spikes = self.snudda_simulation_load.get_spikes(neuron_id=neuron_id)

            for spike in spikes.flatten():
                p_idx = int(spike / period)
                spike_in_period[neuron_id, p_idx] += 1

        for post_id in post_id_list:
            synapses, _ = self.snudda_load.find_synapses(post_id=post_id)
            pre_id_lookup[post_id] = np.array(sorted(list(set(synapses[:, 0]))))

        for post_id in post_id_list:

            pre_id = pre_id_lookup[post_id]
            has_pre_spike = np.sum(spike_in_period[pre_id, :], axis=0)
            has_post_spike = spike_in_period[post_id, :]

            voltage = self.snudda_simulation_load.get_voltage(neuron_id=post_id).flatten()

            for t_start, t_end, pre_spikes, post_spikes \
                    in zip(period_start, period_end, has_pre_spike, has_post_spike):

                if time_range and not (time_range[0] <= t_start and t_end <= time_range[1]):
                    continue

                t_idx = np.where(np.logical_and(t_start <= time, time < t_end))[0]

                if len(t_idx) > 2000:

                    print("Tell me why?!")
                    import pdb
                    pdb.set_trace()

                if pre_spikes:
                    if post_spikes:
                        spike_with_pre_spike.append(voltage[t_idx])
                    else:
                        no_spike_with_pre_spike.append(voltage[t_idx])
                else:
                    if post_spikes:
                        spike_no_pre_spike.append(voltage[t_idx])
                    else:
                        no_spike_no_pre_spike.append(voltage[t_idx])

        t_idx = np.where(np.logical_and(period_start[0] <= time, time < period_end[0]))[0]
        t = time[t_idx] - time[t_idx[0]]

        if len(spike_with_pre_spike) > 0:
            spike_with_pre_spike_mat = np.stack(spike_with_pre_spike, axis=1)
            ax.plot(t * 1e3, 1e3 * np.mean(spike_with_pre_spike_mat, axis=1),
                    label=f"{neuron_type} spike, pre-spike (n={spike_with_pre_spike_mat.shape[1]})")

        if len(no_spike_with_pre_spike) > 0:
            no_spike_with_pre_spike_mat = np.stack(no_spike_with_pre_spike, axis=1)
            ax.plot(t * 1e3, 1e3 * np.mean(no_spike_with_pre_spike_mat, axis=1),
                    label=f"No {neuron_type} spike, pre-spike (n={no_spike_with_pre_spike_mat.shape[1]})")

        if len(spike_no_pre_spike) > 0:
            spike_no_pre_spike_mat = np.stack(spike_no_pre_spike, axis=1)
            ax.plot(t * 1e3, 1e3 * np.mean(spike_no_pre_spike_mat, axis=1),
                    label=f"{neuron_type} spike, no pre-spike (n={spike_no_pre_spike_mat.shape[1]})")

        if len(no_spike_no_pre_spike) > 0:
            no_spike_no_pre_spike_mat = np.stack(no_spike_no_pre_spike, axis=1)
            ax.plot(t*1e3, 1e3*np.mean(no_spike_no_pre_spike_mat, axis=1),
                    label=f"No {neuron_type} spike, no pre-spike (n={no_spike_no_pre_spike_mat.shape[1]})")

        ax.legend()
        plt.xlabel("Time (ms)")
        plt.ylabel("Voltage (mV)")

        if figure_name:
            fig_path = os.path.join(self.figure_path, figure_name)
            plt.savefig(fig_path)

        if show_plot:
            plt.ion()
            plt.show()

        # if show_pre_histogram:
        #     plt.figure(figsize=fig_size)

        return ax



import os
import numpy as np

from snudda.neurons.morphology_data import MorphologyData
from snudda.plotting.plot_spike_raster_v2 import SnuddaPlotSpikeRaster2
from snudda.utils.snudda_path import get_snudda_data, snudda_parse_path
from snudda.neurons.neuron_prototype import NeuronPrototype
from snudda.utils.load import SnuddaLoad
import matplotlib.pyplot as plt


class PlotNetwork(object):

    def __init__(self, network, snudda_data=None):

        if os.path.isdir(network):
            network_file = os.path.join(network, "network-synapses.hdf5")
        else:
            network_file = network

        self.network_file = network_file
        self.network_path = os.path.dirname(self.network_file)

        self.sl = SnuddaLoad(self.network_file)
        self.snudda_data = get_snudda_data(snudda_data=snudda_data,
                                           network_path=self.network_path)
        self.prototype_neurons = dict()
        self.extra_axon_cache = dict()

    def close(self):
        self.sl.close()

    def plot(self, plot_axon=True, plot_dendrite=True, plot_synapses=True,
             neuron_id_list=None, filter_synapses_pre_id_list=None,
             title=None, title_pad=None, show_axis=True,
             elev_azim=None, fig_name=None, dpi=600,
             colour_population_unit=False,
             colour_neuron_type=True):

        if type(plot_axon) == bool:
            plot_axon = np.ones((self.sl.data["num_neurons"],), dtype=bool) * plot_axon

        if type(plot_dendrite) == bool:
            plot_dendrite = np.ones((self.sl.data["num_neurons"],), dtype=bool) * plot_dendrite

        assert len(plot_axon) == len(plot_dendrite) == len(self.sl.data["neurons"])

        fig = plt.figure(figsize=(6, 6.5))
        ax = plt.axes(projection='3d')

        if "simulation_origo" in self.sl.data:
            simulation_origo = self.sl.data["simulation_origo"]
        else:
            simulation_origo = np.array([0, 0, 0])

        if "population_unit" in self.sl.data and colour_population_unit:
            population_unit = self.sl.data["population_unit"]
            pop_units = sorted(list(set(population_unit)))
            cmap = plt.get_cmap('tab20', len(pop_units))
            colour_lookup_helper = dict()

            for idx, pu in enumerate(population_unit):
                if pu > 0:
                    colour_lookup_helper[idx] = cmap(pu)
                else:
                    colour_lookup_helper[idx] = 'lightgrey'

            colour_lookup = lambda x: colour_lookup_helper[x]
        else:
            colour_lookup = lambda x: 'black'

        if colour_neuron_type:
            colour_lookup_helper = dict()

            for nid in self.sl.get_neuron_id():
                colour_lookup_helper[nid] = SnuddaPlotSpikeRaster2.get_colours(self.sl.data["neurons"][nid]["type"])

            colour_lookup = lambda x: colour_lookup_helper[x]

        # Plot neurons
        for neuron_info, pa, pd in zip(self.sl.data["neurons"], plot_axon, plot_dendrite):

            if neuron_id_list:
                # We only plot certain neuron_id
                if neuron_info["neuron_id"] not in neuron_id_list:
                    continue

            soma_colour = colour_lookup(neuron_info["neuron_id"])
            neuron = self.load_neuron(neuron_info)

            neuron.plot_neuron(axis=ax,
                               plot_axon=pa,
                               plot_dendrite=pd,
                               soma_colour=soma_colour,
                               axon_colour=None,  #"darksalmon",  #"maroon",
                               dend_colour=None)  #"silver")   # Can also write colours as (0, 0, 0) -- rgb

        # Plot synapses
        if plot_synapses and "synapse_coords" in self.sl.data and self.sl.data["synapse_coords"].size > 0:

            if neuron_id_list:
                post_id = self.sl.data["synapses"][:, 1]

                keep_flag = np.zeros(post_id.shape, dtype=bool)
                for nid in neuron_id_list:
                    keep_idx = np.where(post_id == nid)[0]
                    keep_flag[keep_idx] = True

                if filter_synapses_pre_id_list:
                    pre_id = self.sl.data["synapses"][:, 0]

                    pre_keep_flag = np.zeros(post_id.shape, dtype=bool)
                    for nid in filter_synapses_pre_id_list:
                        keep_idx = np.where(pre_id == nid)[0]
                        pre_keep_flag[keep_idx] = True

                    keep_flag = np.logical_and(keep_flag, pre_keep_flag)

                keep_idx = np.where(keep_flag)[0]

                x = self.sl.data["synapse_coords"][:, 0][keep_idx]
                y = self.sl.data["synapse_coords"][:, 1][keep_idx]
                z = self.sl.data["synapse_coords"][:, 2][keep_idx]

                plt.figtext(0.5, 0.20, f"{keep_idx.size} synapses", ha="center", fontsize=18)

            else:
                x = self.sl.data["synapse_coords"][:, 0]
                y = self.sl.data["synapse_coords"][:, 1]
                z = self.sl.data["synapse_coords"][:, 2]

            ax.scatter(x, y, z, color=(1, 0, 0))

            if neuron_id_list is None:
                plt.figtext(0.5, 0.20, f"{self.sl.data['num_synapses']} synapses", ha="center", fontsize=18)

        if elev_azim:
            ax.view_init(elev_azim[0], elev_azim[1])

        if not show_axis:
            plt.axis("off")

        plt.tight_layout()

        if title is None:
            title = ""

        if title_pad is not None:
            plt.rcParams['axes.titley'] = 0.95  # y is in axes-relative co-ordinates.
            plt.rcParams['axes.titlepad'] = title_pad  # pad is in points...

        plt.title(title, fontsize=18)

        # ax.dist = 8

        self.equal_axis(ax)

        if fig_name is not None:
            fig_path = os.path.join(self.network_path, "figures", fig_name)
            if not os.path.exists(os.path.dirname(fig_path)):
                os.mkdir(os.path.dirname(fig_path))
            plt.savefig(fig_path, dpi=dpi, bbox_inches="tight")
            print(f"Writing figure: {fig_path}")

        plt.ion()
        plt.show()

        return plt, ax

    def load_neuron(self, neuron_info=None, neuron_id=None):

        assert (neuron_info is None) + (neuron_id is None) == 1, "Specify neuron_info or neuron_id"

        if neuron_id is not None:
            print(f"Using id {neuron_id}")
            neuron_info = self.sl.data["neurons"][neuron_id]

        neuron_name = neuron_info["name"]

        if neuron_name not in self.prototype_neurons:
            self.prototype_neurons[neuron_name] = NeuronPrototype(neuron_name=neuron_info["name"],
                                                                  neuron_path=neuron_info["neuron_path"],
                                                                  snudda_data=self.snudda_data,
                                                                  load_morphology=True,
                                                                  virtual_neuron=False)

        morph_path = snudda_parse_path(neuron_info["morphology"], self.snudda_data)
        if os.path.isfile(morph_path):
            morphology_path = morph_path
        else:
            morphology_path = None  # Get morpholog automatically from morphology_key

        neuron = self.prototype_neurons[neuron_name].clone(position=neuron_info["position"],
                                                           rotation=neuron_info["rotation"],
                                                           parameter_key=neuron_info["parameter_key"],
                                                           morphology_key=neuron_info["morphology_key"],
                                                           modulation_key=neuron_info["modulation_key"],
                                                           morphology_path=morphology_path)

        if "extra_axons" in neuron_info:
            for axon_name, axon_info in neuron_info["extra_axons"].items():

                if axon_info["morphology"] not in self.extra_axon_cache:
                    self.extra_axon_cache[axon_info["morphology"]] = MorphologyData(swc_file=axon_info["morphology"],
                                                                                    parent_tree_info=None,
                                                                                    snudda_data=self.snudda_data)

                neuron.add_morphology(swc_file=axon_info["morphology"],
                                      name=axon_name,
                                      position=axon_info["position"],
                                      rotation=axon_info["rotation"],
                                      morphology_data=self.extra_axon_cache[axon_info["morphology"]])

        return neuron

    def plot_populations(self, unmarked_alpha=0.3):

        fig = plt.figure(figsize=(6, 6.5))
        ax = plt.axes(projection='3d')

        assert "population_unit" in self.sl.data

        population_unit = self.sl.data["population_unit"]
        pop_units = sorted(list(set(population_unit)))
        cmap = plt.get_cmap('tab20', len(pop_units)+1)
        neuron_colours = []

        for idx, pu in enumerate(population_unit):
            if pu > 0:
                neuron_colours.append(list(cmap(pu)))
            else:
                neuron_colours.append([0.6, 0.6, 0.6, 1.0])

        neuron_colours = np.array(neuron_colours)
        positions = self.sl.data["neuron_positions"]

        pop_unit_idx = np.where(population_unit > 0)[0]
        unmarked_idx = np.where(population_unit == 0)[0]

        ax.scatter(positions[pop_unit_idx, 0], positions[pop_unit_idx, 1], positions[pop_unit_idx, 2],
                   c=neuron_colours[pop_unit_idx, :], marker='o', s=20, alpha=1)

        ax.scatter(positions[unmarked_idx, 0], positions[unmarked_idx, 1], positions[unmarked_idx, 2],
                   c=neuron_colours[unmarked_idx, :], marker='o', s=20, alpha=unmarked_alpha)

        # ax.scatter(positions[:, 0], positions[:, 1], positions[:, 2], c=neuron_colours, marker='o', s=20,alpha=alphas)

        self.equal_axis(ax)

        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")

        for pop in pop_units:
            print(f"Population unit {pop} has {len(np.where(population_unit == pop)[0])} neurons")

    def equal_axis(self, ax):

        x_min, x_max = ax.get_xlim()
        y_min, y_max = ax.get_ylim()
        z_min, z_max = ax.get_zlim()

        x_mean = (x_min + x_max)/2
        y_mean = (y_min + y_max)/2
        z_mean = (z_min + z_max)/2

        max_half_width = np.max([x_max-x_min, y_max-y_min, z_max-z_min])/2

        ax.set_xlim(x_mean-max_half_width, x_mean+max_half_width)
        ax.set_ylim(y_mean-max_half_width, y_mean+max_half_width)
        ax.set_zlim(z_mean-max_half_width, z_mean+max_half_width)


def snudda_plot_network_cli():

    from argparse import ArgumentParser
    parser = ArgumentParser(description="Plot snudda network from file (hdf5)")
    parser.add_argument("networkFile", help="Network file (hdf5)", type=str)
    parser.add_argument("--neuronID", "--nid", help="List of Neuron ID to show", type=int, default=None, nargs="+",
                        dest="neuron_id_list")
    parser.add_argument("--showAxons", help="Show Axons of neurons", action="store_true")
    parser.add_argument("--showDendrites", help="Show dendrites of neurons", action="store_true")
    parser.add_argument("--showSynapses", help="Show synapses of neurons", action="store_true")
    parser.add_argument("--preID", help="Only show synapses from pre ID neurons",
                        dest="pre_id_list",
                        nargs="+", type=int, default=None)
    parser.add_argument("--wait", action="store_true")
    args = parser.parse_args()

    pn = PlotNetwork(args.networkFile)
    pn.plot(fig_name="network-plot.png", neuron_id_list=args.neuron_id_list,
            plot_axon=args.showAxons, plot_dendrite=args.showDendrites, plot_synapses=args.showSynapses,
            filter_synapses_pre_id_list=args.pre_id_list)

    if args.wait:
        input("Press any key to exit")


if __name__ == "__main__":
    snudda_plot_network_cli()



import os
import numpy as np
from snudda.utils.load import SnuddaLoad
import matplotlib.pyplot as plt


class PlotDensity(object):

    def __init__(self, network):

        if os.path.isdir(network):
            network_file = os.path.join(network, "network-synapses.hdf5")
        else:
            network_file = network

        self.network_file = network_file
        self.network_path = os.path.dirname(self.network_file)

        self.sl = SnuddaLoad(self.network_file)

    def close(self):
        self.sl.close()

    def plot(self, neuron_type, plot_axis, n_bins=5):

        p_axis = {"x": 0, "y": 1, "z": 2}
        assert plot_axis in p_axis, f"plot_axis must be one of {', '.join(p_axis.keys())}"
        neuron_pos = self.sl.data["neuron_positions"]
        cell_id = self.sl.get_neuron_id_of_type(neuron_type)

        fig = plt.figure()
        plt.hist(neuron_pos[cell_id, p_axis[plot_axis]], bins=n_bins)
        plt.title(f"Density of {neuron_type}")
        plt.xlabel(plot_axis)
        plt.ylabel("Count")

        fig_path = os.path.join(self.network_path, "figures")
        if not os.path.exists(fig_path):
            os.mkdir(fig_path)

        plt.savefig(os.path.join(fig_path, f"density-hist-{plot_axis}.png"))
        plt.ion()
        plt.show()


import os
import numpy as np
from snudda.plotting.plot_input_locations import SnuddaPlotInputLocations


class PlotDegeneration:

    def __init__(self, original_network_path, degenerated_network_path):

        self.original_plot = SnuddaPlotInputLocations(network_path=original_network_path)
        self.degenerated_plot = SnuddaPlotInputLocations(network_path=degenerated_network_path)

        self.fig_path = os.path.join(degenerated_network_path, "figures")
        if not os.path.exists(self.fig_path):
            os.mkdir(self.fig_path)

    def plot_neuron(self, neuron_id, figure_size=None, show_internal_synapses=True, hide_axis=False):
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=figure_size)
        ax = fig.add_subplot(111, projection='3d')
        ax = self.original_plot.plot_neuron_inputs(neuron_id=neuron_id,
                                                   neuron_colour=np.array([0.6, 0.6, 0.6]),
                                                   external_colour=np.array([1, 0.5, 0]),
                                                   internal_colour=np.array([0, 0.5, 1]),
                                                   show_internal_synapses=show_internal_synapses,
                                                   ax=ax,
                                                   size=2,
                                                   save_fig=False,
                                                   show_figure=False,
                                                   hide_axis=hide_axis,
                                                   figure_size=figure_size)

        ax = self.degenerated_plot.plot_neuron_inputs(neuron_id=neuron_id,
                                                      neuron_colour=np.array([0, 0, 0]),
                                                      show_internal_synapses=show_internal_synapses,
                                                      ax=ax,
                                                      size=50,
                                                      save_fig=True,
                                                      hide_axis=hide_axis,
                                                      show_figure=True)


def cli():

    import argparse
    parser = argparse.ArgumentParser("plot_degeneration")
    parser.add_argument("original_network_path", help="Path to original network directory")
    parser.add_argument("degenerated_network_path", help="Path to degenerated network directory")
    parser.add_argument("neuron_id", help="Neuron ID to inspect", type=int)
    args = parser.parse_args()

    pd = PlotDegeneration(original_network_path=args.original_network_path,
                          degenerated_network_path=args.degenerated_network_path)
    pd.plot_neuron(neuron_id=args.neuron_id)


if __name__ == "__main__":
    cli()



import numpy as np
from snudda import SnuddaLoad
import matplotlib.pyplot as plt


class PlotDensitySlice:

    def __init__(self, network_file):

        self.sl = SnuddaLoad(network_file)

        self.neuron_positions = self.sl.data["neuronPositions"]
        self.neuron_type_lookup = dict()

        for nt in set([n["type"] for n in self.sl.data["neurons"]]):
            self.neuron_type_lookup[nt] = self.sl.get_neuron_id_of_type(neuron_type=nt)

    def plot_slice(self, neuron_type,
                   x_min=None, x_max=None,
                   y_min=None, y_max=None,
                   z_min=None, z_max=None,
                   projection=None):

        if projection is None:
            projection = "3d"

        n_neurons = self.sl.data["num_neurons"]

        cell_id = self.neuron_type_lookup[neuron_type]
        ok_mask = np.zeros((n_neurons,), dtype=int)
        ok_mask[cell_id] = 1

        if x_min is not None:
            ok_mask = np.logical_and(ok_mask, x_min <= self.neuron_positions[:, 0])

        if x_max is not None:
            ok_mask = np.logical_and(ok_mask, self.neuron_positions[:, 0] <= x_max)

        if y_min is not None:
            ok_mask = np.logical_and(ok_mask, y_min <= self.neuron_positions[:, 1])

        if y_max is not None:
            ok_mask = np.logical_and(ok_mask, self.neuron_positions[:, 1] <= y_max)

        if z_min is not None:
            ok_mask = np.logical_and(ok_mask, z_min <= self.neuron_positions[:, 2])

        if z_max is not None:
            ok_mask = np.logical_and(ok_mask, self.neuron_positions[:, 2] <= z_max)

        # cell_pos = self.neuron_positions[ok_idx, :].copy()
        ok_idx = np.where(ok_mask)[0]
        cell_pos = self.neuron_positions[ok_idx, :].copy()

        print(f"Plotting {np.count_nonzero(ok_mask)} {neuron_type} neurons")

        # for pos in cell_pos:
        #     print(f"xyz: {pos}")

        fig = plt.figure()

        if projection == "yz":
            plt.scatter(cell_pos[:, 1], cell_pos[:, 2])
            plt.axis("equal")
            fig.tight_layout()

        elif projection == "xz":
            plt.scatter(cell_pos[:, 0], cell_pos[:, 2])
            plt.axis("equal")
            fig.tight_layout()

        elif projection == "xy":
            plt.scatter(cell_pos[:, 0], cell_pos[:, 1])
            plt.axis("equal")
            fig.tight_layout()

        elif projection == "3d":
            ax = fig.add_subplot(projection='3d')
            ax.scatter(xs=cell_pos[:, 0], ys=cell_pos[:, 1], zs=cell_pos[:, 2])
            # elev, azim = 0, np.pi/2
            # ax.view_init(elev, azim)
            plt.ion()
        else:
            print(f"Unknown projection: {projection} (use 'xy', 'xz', 'yz' or '3d')")
            return

        plt.show()



import h5py


class RepositionNeurons(object):
    """ Reposition neurons in the network file. """

    def __init__(self, position_file):

        """ Constructor.

        Args:
            position_file : network position file
        """

        self.position_file = position_file
        self.hdf5_file = h5py.File(position_file, "r+")

        assert "synapses" not in self.hdf5_file["network"], "This code should not be used after detection."

    def __del__(self):
        self.close()

    def close(self):

        """ Close file."""

        if self.hdf5_file:
            self.hdf5_file.close()
            self.hdf5_file = None

    def place(self, neuron_id, position, rotation=None, verbose=True):

        """ Reposition neuron neuron_id to position with rotation.

        Args:
            neuron_id : Neuron ID affected
            position : New position of neuron
            rotation : New rotatoin of neuron
            verbose : Print info?
        """

        info = (f"Moving neuron {neuron_id} from {self.hdf5_file['network/neurons/position'][neuron_id]} "
                f"to {position}")
        if rotation is not None:
            info += f" setting rotation to {rotation}"

        if verbose:
            print(info)

        self.hdf5_file["network/neurons/position"][neuron_id] = position

        if rotation is not None:
            self.hdf5_file["network/neurons/rotation"][neuron_id] = rotation.reshape((9,))

    def set_morphology_id(self, neuron_id, morphology_id):

        raise DeprecationWarning("morpology_id is replaced by morphology_key ")

        """ Set morphologyID for neuron with neuron_id (neuron_id = None means all neurons) """
        if neuron_id is not None:
            self.hdf5_file["network/neurons/morphology_id"][neuron_id] = morphology_id
        else:
            self.hdf5_file["network/neurons/morphology_id"][:] = morphology_id

    def set_morphology_key(self, neuron_id, morphology_key):

        """ Set morphologyKey for neuron with neuron_id (neuron_id = None means all neurons) """
        if neuron_id is not None:
            self.hdf5_file["network/neurons/morphology_key"][neuron_id] = morphology_key
        else:
            self.hdf5_file["network/neurons/morphology_key"][:] = morphology_key

    def set_parameter_key(self, neuron_id, parameter_key):

        """ Set parameterKey for neuron with neuron_id (neuron_id = None means all neurons) """

        if neuron_id is not None:
            self.hdf5_file["network/neurons/parameter_key"][neuron_id] = parameter_key
        else:
            self.hdf5_file["network/neurons/parameter_key"][:] = parameter_key

    def set_parameter_id(self, neuron_id, parameter_id):

        raise DeprecationWarning("parameter_id is replaced by parameter_key")

        """ Set parameterID for neuron with neuron_id (neuron_id = None means all neurons) """

        if neuron_id is not None:
            self.hdf5_file["network/neurons/parameter_id"][neuron_id] = parameter_id
        else:
            self.hdf5_file["network/neurons/parameter_id"][:] = parameter_id

    def set_modulation_id(self, neuron_id, modulation_id):

        raise DeprecationWarning("modulation_id is replaced by modulation_key")

        """ Set modulationID for neuron with neuron_id (neuron_id = None means all neurons) """

        if neuron_id is not None:
            self.hdf5_file["network/neurons/modulation_id"][neuron_id] = modulation_id
        else:
            self.hdf5_file["network/neurons/modulation_id"][:] = modulation_id

    def set_modulation_key(self, neuron_id, modulation_key):

        """ Set modulationKey for neuron with neuron_id (neuron_id = None means all neurons) """

        if neuron_id is not None:
            self.hdf5_file["network/neurons/modulation_key"][neuron_id] = modulation_key
        else:
            self.hdf5_file["network/neurons/modulation_key"][:] = modulation_key




translation ={
    "dendrite" : "basal",
    "soma" : "soma",
    "axon" : "axon"
}

re_translation ={
    "dend" : "dendrite",
    "soma" : "soma",
    "axon" : "axon"
}


def translate(word):

    return translation[word]

    
def re_translate(word):

    return re_translation[word]




import numpy as np


class TimeVaryingInput:

    def __init__(self):

        pass

    @staticmethod
    def get_stretch_time(frequency_function, end_time, start_time=0, dt=0.001, check_positive=True, start_at_zero=True):

        """ We want to stretch the time, so that a Poisson process with frequency 1Hz will result in a time varying
            Poisson process with the instantaneous frequency_function from 0 to end_time, with time resolution dt.

            Any frequency values < 0 are treated as 0.

            Args:
                frequency_function : a numpy compatible function that takes an array and returns an array
                end_time : end time of the valid time range
                start_time : start_time that function is defined for
                dt : time resolution of the sampling of the frequency function
                start_at_zero (bool) : If start_at_zero is True, then the frequency function is f(t=0)
                                       at the start of each interval, if False the simulation time is used.

        """

        time = np.arange(start_time, end_time, dt)

        if start_at_zero:
            func_time = np.arange(0, end_time - start_time, dt)
        else:
            func_time = time

        frequency = frequency_function(func_time)

        # If frequency was a scalar, extend it to be that frequency in entire time range
        if frequency.size == 1:
            frequency = np.full(time.shape, frequency)

        if check_positive:
            frequency[frequency <= 0] = 0

        try:
            stretched_time = np.cumsum(frequency*dt) - frequency[0]*dt  # We want stretched time to start at 0
            func = lambda t, stretched_time=stretched_time, time=time: np.interp(t, stretched_time, time)
            stretch_end_time = stretched_time[-1]
        except:
            import traceback
            print(traceback.format_exc())
            import pdb
            pdb.set_trace()

        return func, stretch_end_time

    @staticmethod
    def _poisson_helper(end_time, rng):

        t_diff = -np.log(1.0 - rng.random(int(np.ceil(max(1, end_time)))))  # Generate 1 Hz spiking
        t_spikes = [np.cumsum(t_diff)]

        # Is last spike after end of duration
        while t_spikes[-1][-1] <= end_time:
            t_diff = -np.log(1.0 - rng.random(int(np.ceil(end_time * 0.1))))
            t_spikes.append(t_spikes[-1][-1] + np.cumsum(t_diff))

        # Prune away any spikes after end
        if len(t_spikes[-1]) > 0:
            t_spikes[-1] = t_spikes[-1][t_spikes[-1] <= end_time]

        # Return spike times
        return np.concatenate(t_spikes)

    @staticmethod
    def generate_spikes(frequency_function, end_time, rng, start_time=0, n_spike_trains=1, start_at_zero=True):

        """
            Generates spikes with frequency f(t) where f is specified by frequency_function.

            Args:
                frequency_function : Either a python function, or a string of a function of t, e.g. "sin(2*pi*5*t)"
                end_time (float) : End time of spike train
                start_time (float) : Start time of spike train (default 0)
                n_spike_trains (int) : Number of spike trains, default 1
                start_at_zero (bool) : Is t=0 at the start of the time interval the function is active within?
                                       Default True. If set to False the simulation t is sent directly to the
                                       frequency_function.
        """

        stretch_func, stretched_end_time = TimeVaryingInput.get_stretch_time(frequency_function=frequency_function,
                                                                             start_time=start_time, end_time=end_time,
                                                                             start_at_zero=start_at_zero)
        spike_trains = []

        for idx in range(0, n_spike_trains):
            t_spike = TimeVaryingInput._poisson_helper(end_time=stretched_end_time, rng=rng)
            spike_trains.append(stretch_func(t_spike))  # Stretch the spike times

        return spike_trains

    @staticmethod
    def test_spike_frequency():

        func = lambda t: 5*np.cos(10*2*np.pi*t) + 6 + 10*t
        rng = np.random.default_rng()

        spikes = TimeVaryingInput.generate_spikes(frequency_function=func, start_time=1, end_time=3,
                                                  n_spike_trains=10000, rng=rng)

        all_spikes = np.concatenate(spikes)

        import matplotlib.pyplot as plt
        plt.figure()

        # We generate 3 seconds of data, 1000 trains, 3000 seconds of total data
        # we place that in 1000 bins, so if we weight result by 1/3 we should show approximate instantaneous frequency
        plt.hist(all_spikes, bins=1000, weights=np.full(all_spikes.shape, 1/20))
        plt.xlabel("Time (s)")
        plt.ylabel("Frequency (Hz)")
        plt.title("Oscillating between 1 and 11 Hz, periodicity 0.1s")
        plt.ion()
        plt.show()
        plt.pause(0.1)


if __name__ == "__main__":

    TimeVaryingInput.test_spike_frequency()
    import pdb
    pdb.set_trace()





#!/usr/bin/env python3

# This code allows the user to remove a portion of the network by cutting a
# slice of the data.

# 0. Load the network
# 1. Define the cut plane
# 2. Find if any of the neurons are above the cut plane, remove those
# 3. Remove any synapses above the cut plane
# 4. Write a new hdf5
#

import os
import sys

import h5py
import numexpr
import numpy as np


# TODO: Check that no population unit list needs to be remapped also, this is old code that was just refactored...


class SnuddaCut(object):
    """ Cuts part of the volume, to simulate creating a slice of tissue, or other cut. """

    def __init__(self, network_file, cut_equation="z>0",
                 out_file_name=None,
                 plot_only=False, show_plot=True):

        """ Constructor.

        Args:
            network_file (str): Path to network file
            cut_equation (str): Cut equation, e.g. "z>0" the neurons fullfilling equation are kept
            out_file_name (str): Path to new network file
            plot_only (bool): Only create a plot of cut
            show_plot (bool): Show plot on screen
        """

        self.cut_equation = cut_equation
        self.h5libver = "latest"
        self.h5driver = "sec2"
        self.in_file = None
        self.out_file = None

        if out_file_name is None:
            base_path = os.path.dirname(network_file)
            self.out_file_name = os.path.join(base_path, "network-cut-slice.hdf5")
        else:
            self.out_file_name = out_file_name

        # We create a lambda expression, to avoid calling eval each time
        self.cut_equation_lambda = lambda x, y, z: numexpr.evaluate(cut_equation)

        self.open_input_file(network_file)

        # TODO: Move these out of init, and into separate function to allow more manipulations
        if plot_only:
            self.plot_cut(include_synapses=False, include_gap_junctions=False)
            sys.exit(0)

        self.setup_output_file(self.out_file_name)
        self.write_cut_slice(self.cut_equation_lambda, self.out_file_name, print_remapping=False)

        self.plot_cut(include_synapses=True, include_gap_junctions=True,
                      show_plot=show_plot)

        if False:
            self.in_file.close()
            self.out_file.close()

    ############################################################################

    # TODO: Split this function, so that one writes slice, but requires info about what to write
    #       while other function determines what we should keep/remove

    def write_cut_slice(self, cut_equation_lambda, out_file_name=None, print_remapping=False):

        """ Write cut slice to file.

        Args:
            cut_equation_lambda : lamba function with cut equation
        """

        """ Write network to hdf5 file: output_file_name """

        # Remove the neurons from the data
        soma_keep_flag = self.soma_inside(cut_equation_lambda)
        soma_keep_id = np.where(soma_keep_flag)[0]
        soma_remove_id = np.where(soma_keep_flag == False)[0]
        num_soma_keep = np.sum(soma_keep_flag)

        if num_soma_keep == 0:
            print("No somas left, aborting!")
            sys.exit(-1)

        print(f"Keeping {num_soma_keep} out of {len(soma_keep_flag)}"
              " neurons (the others have soma outside of cut plane)")

        # We need to remap neuron_id in the synapses and gap junction matrix
        remapping_file = f"{out_file_name}-remapping.txt"
        remap_id = dict([])
        with open(remapping_file, "w") as f:
            for new_id, old_id in enumerate(soma_keep_id):
                remap_id[old_id] = new_id
                f.write(f"{old_id}, {new_id}\n")
            
        if print_remapping:
            print("\nRemapping neurons:")
            for new_id, old_id in enumerate(soma_keep_id):
                assert remap_id[old_id] == new_id, f"Internal error with remap_id"
                if old_id == new_id:
                    print(f"{old_id} the same")
                else:
                    print(f"{old_id} -> {new_id}")
            print("")



        # TODO: Write unit test, that takes a connection matrix, notes how some of the cells are connected
        #       then does cut / ablation, and then verifies that those cells are still connected the same way, if left
        #       + extra unit test, som kollar att inga nya variabler lagts till i load, då vill vi få en krasch
        #       och varning

        network_group = self.out_file.create_group("network")
        neuron_group = network_group.create_group("neurons")

        if len(self.in_file["network/neurons/extra_axons/parent_neuron"][()]) > 0:
            # To implement this we need to only keep the axons that have parent neurons that are still there
            raise NotImplemented("extra_axons currently not supported by cut.py")

        for var_name in self.in_file["network/neurons"]:

            data = self.in_file[f"network/neurons/{var_name}"]

            if type(data) == h5py._hl.group.Group:
                self.in_file.copy(f"network/neurons/{var_name}", neuron_group)
                continue

            if len(data.shape) == 0:
                # Scalar data, just copy
                self.in_file.copy(f"network/neurons/{var_name}", neuron_group)
                continue

            elif len(data.shape) == 1:
                # 1D data, we only keep nSomaKeep of them
                data_shape = (num_soma_keep,)
            elif len(data.shape) == 2:
                # 2D data, need to make sure to maintain dimensions
                data_shape = (num_soma_keep, data.shape[1])
            else:
                print("writeCutSlice: Only handle 0D, 1D and 2D data, update code!")
                sys.exit(-1)

            if var_name == "neuron_id":
                # We need to remap
                neuron_group.create_dataset(var_name, data_shape, data.dtype,
                                            [remap_id[data[x]] for x in soma_keep_id],
                                            compression=data.compression)

                # Double check that it is OK, should be in order after
                assert (np.diff(neuron_group["neuron_id"][()]) == 1).all(), \
                    "Problem with neuron remapping!"

            else:
                try:
                    neuron_group.create_dataset(var_name, data_shape, data.dtype,
                                                [data[x] for x in soma_keep_id],
                                                compression=data.compression)
                except:
                    import traceback
                    tstr = traceback.format_exc()
                    print(tstr)
                    sys.exit(-1)

        if "synapses" in self.in_file["network"]:

            # Next deal with synapses
            keep_syn_flag = self.synapses_inside(cut_equation_lambda, data_type="synapses")
            keep_syn_flag = self.filter_neurons_synapses(soma_remove_id,
                                                         keep_flag=keep_syn_flag,
                                                         data_type="synapses")

            # Lastly deal with gap junctions
            keep_gj_flag = self.synapses_inside(cut_equation_lambda, data_type="gap_junctions")
            keep_gj_flag = self.filter_neurons_synapses(soma_remove_id,
                                                        keep_flag=keep_gj_flag,
                                                        data_type="gap_junctions")

            num_syn = np.sum(keep_syn_flag)
            num_synapses = np.zeros((1,), dtype=np.uint64) + num_syn

            num_gj = np.sum(keep_gj_flag)
            num_gap_junctions = np.zeros((1,), dtype=np.uint64) + num_gj

            network_group.create_dataset("num_synapses", data=num_synapses, dtype=np.uint64)
            network_group.create_dataset("num_gap_junctions", data=num_gap_junctions,
                                         dtype=np.uint64)

            syn_mat = self.in_file["network/synapses"][()].copy()
            gj_mat = self.in_file["network/gap_junctions"][()].copy()

            print("Copying synapses and gap junctions")

            for idx, row_idx in enumerate(np.where(keep_syn_flag)[0]):
                # We need to remap the neuron_id if some neurons have been removed!!
                syn_mat[row_idx, 0] = remap_id[syn_mat[row_idx, 0]]
                syn_mat[row_idx, 1] = remap_id[syn_mat[row_idx, 1]]

            network_group.create_dataset("synapses", 
                                         data=syn_mat[np.where(keep_syn_flag)[0], :],
                                         dtype=np.int32, 
                                         compression="gzip")

            print(f"Keeping {num_syn} synapses (out of {syn_mat.shape[0]})")

            for idx, row_idx in enumerate(np.where(keep_gj_flag)[0]):
                # We need to remap the neuron_id if some neurons have been removed!!
                gj_mat[row_idx, 0] = remap_id[gj_mat[row_idx, 0]]
                gj_mat[row_idx, 1] = remap_id[gj_mat[row_idx, 1]]
                
            network_group.create_dataset("gap_junctions",
                                         data=gj_mat[np.where(keep_gj_flag)[0], :],
                                         dtype=np.int32, 
                                         compression="gzip")

            print(f"Keeping {num_gj}  gap junctions (out of {gj_mat.shape[0]})")

        else:
            print("No synapses found (assuming this was a save file with only position information).")

    ############################################################################

    # Tells which somas are inside the cut equation

    def soma_inside(self, cut_equation_lambda):

        """ Check if soma are inside cut_equation_lambda. Returns a boolean numpy array. """

        pos = self.in_file["network/neurons/position"][()]
        inside_flag = np.array([cut_equation_lambda(x, y, z) for x, y, z in pos], dtype=bool)

        return inside_flag

    ############################################################################

    def synapses_inside(self, cut_equation_lambda, data_type="synapses"):

        """ Check if synapses are inside cut_equation_lambda. Returns a numpy bool array.

        Args:
            cut_equation_lambda : lambda function representing cut
            data_type : e.g. 'synapses' or 'gap_junctions'
        """

        voxel_size = self.in_file["meta/voxel_size"][()]
        sim_origo = self.in_file["meta/simulation_origo"][()]

        if data_type == "synapses":
            pos = self.in_file["network/synapses"][:, 2:5] * voxel_size + sim_origo
        elif data_type == "gap_junctions":
            pos = self.in_file["network/gap_junctions"][:, 6:9] * voxel_size + sim_origo
        else:
            print(f"filterNeuronsSynapses: Unknown data_type: {data_type} (valid are 'synapses' or 'gap_junctions'")
            sys.exit(-1)

        inside_flag = np.array([cut_equation_lambda(x, y, z) for x, y, z in pos], dtype=bool)

        return inside_flag

    ############################################################################

    # Returns the row numbers that do not contain the neuron_id, ie filters
    # the synapses belonging to neuron_id out...

    # dataType = "synapses" or "gap_junctions"

    def filter_neurons_synapses(self, neuron_id, keep_flag=None, data_type="synapses"):

        """ Filter synapse matrix, to only keep those synapses that belong to neuron_id.

        Args:
            neuron_id : Neuron ID to remove
            keep_flag : Which synapses are available to pick from
            data_type : "synapses" or "gap_junctions"

        Returns:
            keep_flag : bool array with which synapses to keep
        """
        print(f"filtering {data_type}")
        num_neurons = self.in_file["network/neurons/neuron_id"].shape[0]
        neuron_keep_flag = np.ones((num_neurons,),dtype=bool)
        for nid in neuron_id:
            neuron_keep_flag[nid] = False

        if data_type == "synapses":
            data_str = "network/synapses"
        elif data_type == "gap_junctions":
            data_str = "network/gap_junctions"
        else:
            print(f"filter_neurons_synapses: Unknown data_type: {data_type} (valid are 'synapses', 'gap_junctions'")
            sys.exit(-1)

        if keep_flag is None:
            keep_flag = np.ones((self.in_file[data_str].shape[0],), dtype=bool)

        source_id = self.in_file[data_str][:, 0]
        destination_id = self.in_file[data_str][:, 1]

        for idx, (src_id, dest_id) in enumerate(zip(source_id, destination_id)):
            keep_flag[idx] = neuron_keep_flag[src_id] and neuron_keep_flag[dest_id]

        print(f"filtering of {data_type} done")
        return keep_flag

    ############################################################################

    def open_input_file(self, network_file):

        """ Opens original network_file"""

        self.in_file = h5py.File(network_file, "r", libver=self.h5libver, driver=self.h5driver)

    ############################################################################

    # This sets up the output file, copies the config and meta data,
    # but does not copy over the neurons, synapses or gap junctions

    def setup_output_file(self, out_file_name):

        """ Creates output network file, out_file_name """

        print(f"Writing to {out_file_name}")

        self.out_file = h5py.File(out_file_name, "w", libver=self.h5libver, driver=self.h5driver)

        print("Copying 'config' and 'meta'")
        if "config" in self.out_file:
            self.in_file.copy("config", self.out_file)

        self.in_file.copy("config", self.out_file)
        self.in_file.copy("meta", self.out_file)

        if "morphologies" in self.in_file:
            print("Copying morphologies")
            self.in_file.copy("morphologies", self.out_file)

    ############################################################################

    # This is just used to verify

    def plot_cut(self, include_synapses=True, include_gap_junctions=True, show_plot=True):

        """ Plot the cut to verify it is what we want.

        Args:
            include_synapses (bool) : Plot synapses?
            include_gap_junctions (bool) : Plot gap junctions?
            show_plot (bool) : Plot, or just write to file?
        """

        if "voxel_size" not in self.in_file["meta"]:
            print("plot_cut currently works after detect has been done, not plotting place files.")
            return

        print("Plotting verification figure")

        voxel_size = self.in_file["meta/voxel_size"][()]
        sim_origo = self.in_file["meta/simulation_origo"][()]

        in_pos = self.in_file["network/neurons/position"][()]
        in_syn = self.in_file["network/synapses"][:, 2:5] * voxel_size + sim_origo
        in_gj = self.in_file["network/gap_junctions"][:, 6:9] * voxel_size + sim_origo

        if self.out_file is not None:
            # Just double check that they match
            assert self.out_file["meta/voxel_size"][()] == voxel_size
            assert (self.out_file["meta/simulation_origo"][()] == sim_origo).all()

            out_pos = self.out_file["network/neurons/position"][()]
            out_syn = self.out_file["network/synapses"][:, 2:5] * voxel_size + sim_origo
            out_gj = self.out_file["network/gap_junctions"][:, 6:9] * voxel_size + sim_origo

        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        a_val = 0.5
        a_val2 = 0.2

        ax.scatter(in_pos[:, 0], in_pos[:, 1], in_pos[:, 2], c='black', s=25, alpha=a_val2)
        if self.out_file is not None:
            ax.scatter(out_pos[:, 0], out_pos[:, 1], out_pos[:, 2], c='red', s=15, alpha=a_val)

        if include_synapses:
            ax.scatter(in_syn[:, 0], in_syn[:, 1], in_syn[:, 2], c='black', s=9, alpha=a_val2)

            if self.out_file is not None:
                ax.scatter(out_syn[:, 0], out_syn[:, 1], out_syn[:, 2], c='red', s=4, alpha=a_val)

        if include_gap_junctions:
            ax.scatter(in_gj[:, 0], in_gj[:, 1], in_gj[:, 2], c='blue', s=6, alpha=a_val2)
            if self.out_file is not None:
                ax.scatter(out_gj[:, 0], out_gj[:, 1], out_gj[:, 2], c='green', s=3, alpha=a_val)

        ax.view_init(elev=-3, azim=-95)

        fig_name = f"{self.out_file_name}.pdf"
        print(f"Writing to figure {fig_name}")
        plt.savefig(fig_name, dpi=300)

        if show_plot:
            plt.show()
            print("close figure to exit program!")

    ############################################################################


if __name__ == "__main__":

    from argparse import ArgumentParser

    parser = ArgumentParser(description="Cut a slice from a network")
    parser.add_argument("networkFile", help="Network file (hdf5)", type=str)
    parser.add_argument("cutEquation",
                        help="Equation defining parts left after cut, "
                             + "e.g. 'z>0' or 'x+y>100e-6' (SI units)",
                        type=str)
    parser.add_argument("--plotOnly",
                        help="Plots the network without cutting",
                        action="store_true")
    parser.add_argument("--hidePlot", help="Hide plot.", action="store_true")

    args = parser.parse_args()

    if args.hidePlot:
        showPlot = False
    else:
        showPlot = True

    sc = SnuddaCut(network_file=args.networkFile,
                   cut_equation=args.cutEquation,
                   plot_only=args.plotOnly,
                   show_plot=showPlot)

